#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **expression alias** (i.e., objects satisfying the data descriptor
protocol dynamically defined at class scope) facilities.
'''

#FIXME: Cache the values to which expressions evaluate in the data descriptor
#classes dynamically created by the following functions *AFTER* either:
#
#* The first call to the special __get__() method successfully validates the
#  expression associated with the passed instance to be valid.
#* Any call to the special __set__() method successfully validates the passed
#  new value associated with the passed instance to be valid.
#
#This is quite critical, as these method bodies are becoming increasingly
#lengthy and hence computationally expensive. Maybe you are now thinking to
#yourself: "If it's so critical, why haven't you done this already?" Because
#data descriptors are per-*CLASS* whereas expression values are per-*OBJECT*.
#Mapping between the two requires heavy lifting, which (because the whole point
#of caching is efficiency) must necessarily be as speedy as feasible.
#
#Two approaches present themselves:
#
#* Define the special __set_name__() method in these data descriptor classes to
#  something resembling:
#
#    def __set_name__(self_descriptor, self, var_name)
#        cache_var_name = '__betsee_cache_' + var_name
#
#        if hasattr(self, cache_var_name):
#            raise SomeBetseException(
#                'Expression alias cache variable "{}" already defined.'.format(
#                    cache_var_name)
#
#  *wAIT.* That actually fails to help us at all (and, in any case, is only
#  supported under Python 3.6), implying that the only remaining option is the
#  following general-purpose approach.
#* Alternately, define a new "__expr_alias_cache" dictionary in *EVERY* object
#  defining one or more expression aliases, mapping from the string value of the
#  "expr" parameter for that alias to the currently cached value of that alias.
#  This in turn implies that the string value of the "expr" parameter for that
#  alias will now need to be passed to the __init__() method defined for that
#  alias and then classified as an instance variable of that alias... *WAIT.*
#  Nope. Since the "expr" parameter never changes for a given alias, the lookup
#  into the "__expr_alias_cache" dictionary can simply be hard-coded into the
#  class definition at expression alias creation time: e.g.,
#
#      '''
#      # Cache this expression's current value.
#      self.__expr_alias_cache[{expr_gettable!r}] = value
#      '''
#
#  The only critical point is the usage of the "!r" flag to ensure this
#  expression's expansion as a proper string. Otherwise, this is dead simple.
#
#O.K., the latter approach is clearly superior in every possible way, so that's
#the way forward. Naturally, we'll need to *EXTENSIVELY* test such caching once
#implemented -- but that should pose no significant burden.
#FIXME: Note that not all expressions are safely cachable -- only those for
#which the underlying expression is guaranteed *NOT* to change between calls to
#the __get__() method. Equivalently, the only safely cachable expression aliases
#are those for which the underlying expression is *NEVER* modified directly but
#only ever through those aliases. To support this concept of safety, a new
#optional "is_cachable" parameter defaulting to "False" will need to be
#accepted by the following methods.

#FIXME: If optimizations are enabled (i.e., "if not __debug__:"), avoid
#embedding type validation in the data descriptor classes defined below.

# ....................{ IMPORTS                           }....................
from betse.exceptions import (
    BetseExprAliasException, BetseEnumException, BetseTypeException)
from betse.util.type.types import (
    type_check,
    BoolOrNoneTypes,
    CallableOrNoneTypes,
    EnumType,
    SequenceTypes,
    StrOrNoneTypes,
    TestableOrNoneTypes,
)

if False: BetseTypeException   # squelch IDE warnings

# ....................{ GLOBALS                           }....................
_EXPR_ALIAS_ID = 0
'''
Unique arbitrary identifier with which to uniquify the class name of the next
:func:`expr_alias` descriptor.
'''

# ....................{ DESCRIPTORS                       }....................
@type_check
def expr_alias(
    # Mandatory parameters.
    expr: str,

    # Optional parameters. For caller convenience, the second parameter passed
    # to this function is required to be the "cls" parameter; doing so permits
    # callers to concisely call this function with positional parameters only.
    cls: TestableOrNoneTypes = None,
    base_classes: SequenceTypes = (),
    expr_settable: StrOrNoneTypes = None,
    is_castable: BoolOrNoneTypes = None,
    predicate: CallableOrNoneTypes = None,
    predicate_expr: StrOrNoneTypes = None,
    predicate_label: StrOrNoneTypes = None,
    obj_name: StrOrNoneTypes = 'self',
) -> object:
    '''
    Expression alias **data descriptor** (i.e., object satisfying the data
    descriptor protocol, usually defined at class scope), dynamically aliasing
    a target variable of the passed type and/or satisfying the passed predicate
    bound to instances of the class subclassing the passed base classes and
    instantiating this descriptor to an arbitrarily complex source Python
    expression suitable for use as both the left- and right-hand sides of
    Python assignment statements.

    This function (in order):

    #. Dynamically defines a new descriptor class specific to the passed string
       such that:

       * Getting the value of the target variable internally gets the current
         value of the source expression.
       * Setting the value of the target variable either:

         * If this value is of the expected type, internally sets the current
           value of the source expression to this value.
         * Else, raises a type exception.

       * Deleting the target variable deletes this variable from the instance
         to which this variable is bound much as for a typical instance
         variable.  Deleting this variable does *not* delete the variable bound
         to the source expression.

    #. Creates an instance of this descriptor class.
    #. Returns this instance.

    The typical use case for this class is to bind an instance variable to a
    nested entry of the dictionary produced by loading a YAML-formatted file.
    Doing so implicitly propagates modifications to this variable back to this
    dictionary and thus back to this file on saving this dictionary in YAML
    format back to this file.

    Expressions
    ----------
    The passed expression may be any arbitrarily complex source Python
    expression suitable for use as both the left- and right-hand sides of
    Python assignment statements. This expression may refer to the following
    local variables:

    * ``self``, the current instance of the class instantiating this
      descriptor. Callers preferring an alternate name for this local variable
      may pass the optional ``obj_name`` parameter. If this descriptor is
      retrieved:
      * As an **instance variable** (i.e., from the current instance of this
        class), this local is that instance.
      * As a **class variable** (i.e., from this class), this local is `None`.
        In this case, the ``cls`` local is guaranteed to be defined to the
        class instantiating this descriptor.
    * ``cls``, the class instantiating this descriptor. This local is only
      defined when getting this descriptor as a class rather than instance
      variable; equivalently, this local is undefined both when getting this
      descriptor as an instance variable *and* when setting this descriptor.
      Since one or the other of this and the ``self`` local is *always*
      defined, the class instantiating this descriptor is safely retrievable in
      expressions by referencing ``cls`` only if ``self`` is ``None``. For
      example, to get and set the name of this class:

      .. code:: python

         expr_alias(
             expr='(cls if self is None else self.__class__).__name__', cls=str)

    * ``self_descriptor``, the current instance of this descriptor. Since this
      descriptor only implements special methods required by the data
      descriptor protocol (e.g., ``__get__``, ``__set__``) and hence contains
      no data, this local is unlikely to be of interest to most callers.

    Class
    ----------
    The returned class is a dynamically synthesized subclass of all passed
    subclasses if any *or* of :class:`object` otherwise. This subclass exposes
    the following public instance variables:

    expr_alias_cls : ClassOrNoneTypes
        Either:

        * If the ``cls`` parameter passed to this method is non-``None``, the
          value of this parameter (i.e., the class or tuple of classes that the
          value of this expression is required to be an instance of).
        * Else, ``None``.

    Caveats
    ----------
    As with all descriptors, this function is intended to be called *only* at
    **class scope** (i.e., directly from within the body of a class) in a
    manner assigning the descriptor returned by this function to a class
    variable of arbitrary name. While feasible, calling this function from any
    other scope is highly discouraged. In turn, this implies this function is
    intended to be called *only* when the expression to be aliased is
    statically known at class definition time. When this is *not* the case
    (e.g., when this expression is only dynamically known at runtime), the
    higher-level :class:`ExprAlias` class should be instantiated instead.

    Additionally, the instance variable bound by the descriptor returned by
    this function is *not* deletable via the :func:`del` builtin. Attempting to
    do so reliably raises an :class:`AttributeError` exception. Why? To
    preserve backward compatibility with pre-Python 3.6 versions, which provide
    no means of doing so.

    Parameters
    ----------
    expr : str
        Arbitrarily complex Python expression suitable for use as at least the
        right-hand side of Python assignment statements, typically
        evaluating to the value of an arbitrarily nested dictionary key:

        * Prefixed by the name of the instance variable to which this
          dictionary is bound in instances of the class containing this
          descriptor (e.g., ``self._config``).
        * Suffixed by one or more ``[``- and ``]``-delimited key lookups into
          the same dictionary (e.g.,
          ``['variable settings']['noise']['dynamic noise']``).

        If this expression is unsuitable for use as the left-hand side of
        Python assignment statements (e.g., due to expanding to a complex
        expression comprising multiple variables rather than to a single simple
        variable), the ``expr_settable`` parameter *must* be passed as well.
    expr_settable : optional[str]
        Arbitrarily complex Python expression suitable for use as the left-hand
        side of Python assignment statements, creating a data descriptor whose:

        * ``__get__()`` implementation embeds the value of the passed ``expr``
          parameter.
        * ``__set__()`` implementation embeds the value of this parameter.

        Defaults to ``None``, in which case this parameter defaults to the
        value of the ``expr`` parameter.
    base_classes : optional[SequenceTypes]
        Sequence of all base classes of the class dynamically synthesized by
        this function for the returned expression alias data descriptor,
        usually used to unify all such descriptors under a common abstract base
        class.  Defaults to the empty tuple, equivalent to the 1-tuple
        ``(object,)`` containing only the root base class of all classes.
    cls : optional[TestableTypes]
        Either:

        * A class. If the value of this expression is not an instance of this
          class *and*:

          * If this value is **safely castable** (i.e., convertible *without*
            raising an exception) into an instance of this class, this value is
            silently casted into an instance of this class by instantiating
            this class with a single positional argument whose value is this
            value (e.g., ``float('3.1415')`` to cast the string value
            ``3.1415`` into a floating point number).
          * Else, an exception is raised.

        * A tuple of classes, in which case an exception is raised if the value
          of this expression is *not* an instance of at least one class in this
          tuple.
        Defaults to ``None``, in which case no such validation is performed.
    is_castable : optional[bool]
        ``True`` only if the value of this expression is safely castable into
        an instance of the class passed as the ``cls`` parameter. If:

        * ``True``, an exception is raised at evaluation time only if this
          value is *not* safely castable into an instance of this class.
        * ``False``, an exception is raised at evaluation time if this value is
          *not* already an instance of this class.

        Defaults to ``None``, in which case this boolean conditionally defaults
        to ``True`` only if the ``cls`` parameter is :class:`float`, whose
        constructor is well-known to safely cast into most builtin scalar types
        (e.g., :class:`int`, :class:`str`).
    predicate : optional[CallableTypes]
        Callable passed the value of this expression as its first and only
        parameter and returning a boolean ``True`` only if this value satisfies
        arbitrary caller requirements and ``False`` otherwise. For uniformity,
        this callable should ideally return boolean values rather than raise
        type exceptions. If this callable returns ``False``, an exception is
        raised on behalf of this callable. If the ``predicate_label``
        parameter is *not* also passed, an exception is raised. Defaults to
        ``None``, in which case no such validation is performed.
    predicate_expr : optional[str]
        Arbitrarily complex Python expression suitable for use as the condition
        of an if statement testing the local variable named ``value`` providing
        the value of this expression evaluating to a boolean ``True`` only if
        the value of this expression satisfies arbitrary caller requirements
        and ``False`` otherwise. If this expression evaluates to ``False``, an
        exception is raised on behalf of this expression. If the
        ``predicate_label`` parameter is *not* also passed, an exception is
        raised. Defaults to ``None``, in which case no such validation is
        performed.
    predicate_label : optional[str]
        Human-readable adjective or adjectival phrase describing the passed
        callable predicate if any. Defaults to ``None``, in which case this
        string is synthesized from the name of this callable if needed.
    obj_name : optional[str]
        Name of the variable in this expression whose value is the current
        instance of the class instantiating this descriptor. If this name is
        already reserved for internal use by this descriptor, an exception is
        raised; reserved names include ``cls``, ``self_descriptor``, and
        ``value``. Defaults to ``self``.

    Returns
    ----------
    base_classes
        Expression alias data descriptor as detailed above, guaranteed to be an
        instance of all passed base classes.

    See Also
    ----------
    :class:`betse.util.type.descriptor.datadesc`
        Submodule providing higher-level classes encapsulating data descriptors
        returned by this lower-level function, intended for use where declaring
        a descriptor at class scope is inappropriate (e.g., where the
        expression to be aliased is unknown at class definition time).
    '''

    # Avoid circular import dependencies.
    from betse.util.type.cls import classes

    # Set of the names of all parameters hard-coded into the implementations of
    # either the __get__() or __set__() methods defined below and hence
    # reserved for internal use by this descriptor.
    RESERVED_ARG_NAMES = {'cls', 'self_descriptor', 'value',}

    # If the passed object name is already reserved, raise an exception.
    if obj_name in RESERVED_ARG_NAMES:
        raise BetseExprAliasException(
            'Object name "{}" reserved for internal use.'.format(obj_name))

    # Dictionary eventually containing only the following three keys:
    #
    # * "__init__", whose value is the __init__() special method for this
    #   descriptor class defined below.
    # * "__get__", whose value is the __get__() special method for this
    #   descriptor class defined below.
    # * "__set__", whose value is the __set__() special method for this
    #   descriptor class defined below.
    class_method_name_to_func = {
        # Passed validational class if any.
        '__expr_alias_cls': cls,

        # Passed validational predicate if any.
        '__expr_alias_predicate': predicate,
    }

    # Expression to be embedded in this data descriptor's __get__() method.
    expr_gettable = expr

    # Expression to be embedded in this data descriptor's __set__() method.
    if expr_settable is None:
        expr_settable = expr_gettable

    # True only if the value to which this expression evaluates is castable.
    if is_castable is None:
        is_castable = cls is float

    # Expression to be embedded in human-readable single-quoted exception
    # messages raised by this data descriptor's __get__() method. To permit
    # this expression to be embedded in single-quoted strings, all single
    # quotes in this expression are globally escaped for safety.
    expr_gettable_single_quotable = expr_gettable.replace("'", "\\'")

    # Python code snippet listing all optional arguments to be accepted by this
    # data descriptor's __init__() method.
    class_init_args = ''

    # Python code snippet implementing the body of this data descriptor's
    # __init__(), __get__(), and __set__() methods respectively.
    class_init_body = ''
    class_get_body = ''
    class_set_body = ''

    # Python code snippet validating the value this expression evaluates to.
    value_test_block = ''

    # If no validational class was passed...
    if cls is None:
        # Nullify the instance variable storing this class for safety.
        class_init_body += '''
    self_descriptor.expr_alias_cls = None'''
    # Else, a validational class was passed. In this case...
    else:
        # Pass this class to this method.
        class_init_args += ', __expr_alias_cls=__expr_alias_cls'

        # Classify this passed class in this method.
        class_init_body += '''
    self_descriptor.expr_alias_cls = __expr_alias_cls'''

        # If the value to which this expression evaluates is castable, attempt
        # to do so and, in the event of failure, wrap the typically
        # non-human-readable low-level exception raised by Python with a
        # human-readable high-level exception.
        #
        # Note that:
        #
        # * The "value" local has been previously defined and hence is
        #   safely accessible here. (See subsequent logic for this definition.)
        # * While this type validation could be performed by decorating the
        #   __get__() and __set__() methods defined below by the @type_check
        #   decorator, doing so would impose overhead for little gain.
        value_test_block += '''
    if not isinstance(value, self_descriptor.expr_alias_cls):'''
        if is_castable:
            value_test_block += '''
        try:
            value = self_descriptor.expr_alias_cls(value)
        except Exception as exception:
            raise BetseTypeException(
                'Expression alias value {!r} not a {!r}.'.format(
                value, self_descriptor.expr_alias_cls)) from exception'''
        # Else, raise an exception unless this value is of the desired type(s).
        else:
            value_test_block += '''
        raise BetseTypeException(
            'Expression alias value {!r} not a {!r}.'.format(
                value, self_descriptor.expr_alias_cls))'''


    # If a predicate was passed...
    if predicate is not None:
        # If no predicate label was passed, raise an exception.
        if predicate_label is None:
            raise BetseExprAliasException(
                'Parameter "predicate_expr" but not "predicate_label" passed.')

        # Pass this predicate to this method.
        class_init_args += ', __expr_alias_predicate=__expr_alias_predicate'

        # Classify this passed predicate in this method.
        class_init_body += '''
    self_descriptor.__expr_alias_predicate = __expr_alias_predicate'''

        # Raise an exception unless the value to which this expression
        # evaluates satisfies the same predicate.
        value_test_block += '''
    if not self_descriptor.__expr_alias_predicate(value):
        raise BetseTypeException(
            'Expression alias value {{!r}} not {predicate_label}.'.format(value))
    '''.format(predicate_label=predicate_label)

    # If a predicate expression was passed...
    if predicate_expr is not None:
        # If no predicate label was passed, raise an exception.
        if predicate_label is None:
            raise BetseExprAliasException(
                'Parameter "predicate_expr" passed, but '
                'parameter "predicate_label" unpassed.')

        # Raise an exception unless the value to which this expression
        # evaluates satisfies the same predicate.
        value_test_block += '''
    if not ({predicate_expr}):
        raise BetseTypeException(
            'Expression alias value {{!r}} not {predicate_label}.'.format(value))
    '''.format(predicate_expr=predicate_expr, predicate_label=predicate_label)

    # If the __init__() method's body remains empty, implement this body as a
    # noop to ensure syntactic validity.
    if not class_init_body:
        class_init_body = '''
    pass'''

    # Prefix all possible implementations of the __get__() method body with a
    # conditional handling the two styles with which Python calls this method
    # -- specifically, class attribute access. This conditional is common
    # boilerplate prefixing most implementations of this method.
    #
    # To quote the official documentation:
    #
    #   "object.__get__(self, instance, owner)
    #
    #    Called to get the attribute of the "owner" class (class attribute
    #    access) or of an instance of that class (instance attribute access).
    #    "owner" is always the owner class, while "instance" is the instance
    #    that the attribute was accessed through, or None when the attribute is
    #    accessed through the owner. This method should return the (computed)
    #    attribute value or raise an AttributeError exception."
    class_get_body = '''
    if {obj_name} is None:
        return self_descriptor
    '''.format(obj_name=obj_name)

    # If this expression is validated...
    if value_test_block:
        # Implement the __get__() method body to validate this expression.
        class_get_body += '''
    # Value to which this expression evaluates to be returned. For readability,
    # wrap all exceptions raised by this evaluation with a human-readable
    # exception exposing this expression alias.
    #
    # While exception handling technically imposes a minor performance penalty,
    # this cost is substantially outweighed by the usability gains. Indeed,
    # "try" blocks are faster in Python than the corresponding "if" statements
    # assuming no exceptions are raised. For detailed timings, see the
    # following authoritative StackOverflow answer:
    #
    #     https://stackoverflow.com/a/2522013/2809027
    try:
        value = {expr_gettable}
    except Exception as exception:
        raise BetseExprAliasException(
            'Expression alias "{expr_gettable_single_quotable}" invalid.'
        ) from exception

    # Validate this value.
    {value_test_block}

    # Return this value.
    return value'''.format(
            expr_gettable=expr_gettable,
            expr_gettable_single_quotable=expr_gettable_single_quotable,
            value_test_block=value_test_block
        )

        # Implement the __set__() method body to validate this expression.
        class_set_body = '''
    # Validate this value.
    {value_test_block}

    # Set this expression to this value.
    {expr_settable} = value'''.format(
            expr_settable=expr_settable,
            value_test_block=value_test_block)
    # Else, this expression is unvalidated. For efficiency, reduce the
    # __get__() and __set__() method bodies to the expected one-liners.
    else:
        class_get_body += '''
    return {expr_gettable}'''.format(expr_gettable=expr_gettable)
        class_set_body = '''
    {expr_settable} = value'''.format(expr_settable=expr_settable)

    #FIXME: Improve exception handling to report the name of the data descriptor
    #in question. To do so, we'll need to implement the data descriptor method
    #passing this data descriptor its name. Only available under Python 3.6,
    #we believe, but that's quite alright. (See below for method signature.)

    # Python code snippet declaring these special methods.
    #
    # Note the differentiation between the "self_descriptor" parameter
    # referring to the current descriptor and the "self" parameter referring to
    # the object containing this descriptor, which may be referenced by the
    # passed expression and must thus be assigned the name "self".
    #
    # Note that the following special methods supported by the descriptor
    # protocol are intentionally left undefined:
    #
    # * "__set_name__(self, owner, name)", first introduced by Python 3.6 to
    #   notify descriptors of the variable names to which they have been bound.
    #   Since these names are irrelevant for aliasing purposes, this descriptor
    #   class leaves this special method unimplemented.
    # * "__delete__(self, instance)", deleting the variable to which this
    #   descriptor has been bound from the current instance of the class
    #   instantiating this descriptor. Sanely defining this method would
    #   require implementing the aforementioned __set_name__() method to store
    #   the name of the variable to which this descriptor is bound. That method
    #   is only available under Python 3.6, whereas this application is
    #   currently compatible with older Python 3.x versions. Since deletion of
    #   this descriptor is non-essential (and arguably undesirable), this
    #   descriptor class leaves this special method unimplemented.
    class_body = '''
def __init__(self_descriptor{init_args}):
    {init_body}

def __get__(self_descriptor, {obj_name}, cls):
    {get_body}

def __set__(self_descriptor, {obj_name}, value):
    {set_body}
'''.format(
    init_args=class_init_args,
    init_body=class_init_body,
    get_body=class_get_body,
    set_body=class_set_body,
    obj_name=obj_name,
)

    # Define these methods and this dictionary containing these methods.
    exec(class_body, globals(), class_method_name_to_func)

    # Prevent input parameters passed into this snippet by this exec() call
    # from polluting the attribute namespace of this class.
    del class_method_name_to_func['__expr_alias_cls']
    del class_method_name_to_func['__expr_alias_predicate']

    # Descriptor class with this name containing only these methods.
    expr_alias_class = classes.define_class(
        base_classes=base_classes,
        class_name=_get_expr_alias_class_name(),
        class_attr_name_to_value=class_method_name_to_func,
    )

    # Instantiate and return the singleton descriptor for this class.
    return expr_alias_class()

# ....................{ DESCRIPTORS ~ enum                }....................
@type_check
def expr_enum_alias(
    # Mandatory parameters.
    expr: str,
    enum_type: EnumType,

    # Optional parameters.
    base_classes: SequenceTypes = (),
) -> object:
    '''
    Enumeration-specific expression alias **data descriptor** (i.e., object
    satisfying the data descriptor protocol), dynamically aliasing a target
    variable of the passed enumeration type bound to instances of the class
    subclassing the passed base classes and instantiating this descriptor to an
    arbitrarily complex source Python expression suitable for use as both the
    left- and right-hand sides of Python assignment statements.

    Invariants
    ----------
    This data descriptor implicitly converts the value to which this expression
    evaluates to and from the corresponding member of this enumeration type in
    a case-insensitive manner. To facilitate this conversion, the caller *must*
    guarantee the following two invariants:

    #. This expression *must* evaluate to a strictly lowercase string value.
       where not the case, an exception will be explicitly raised.
    #. The names of all members of this enumeration type *must* be strictly
       uppercase. Where not the case (i.e., where members have either lowercase
       or mixed case names), an exception will be explicitly raised.

    If these invariants hold, then:

    * Getting the value of this data descriptor implicitly converts the string
      to which this expression evaluates into the member of this enumeration
      whose name is that string uppercased.
    * Setting the value of this data descriptor to a member of this enumeration
      implicitly sets the string to which this expression evaluates to the
      lowercased name of this member.

    Class
    ----------
    The returned class is a dynamically synthesized subclass of all passed
    subclasses if any *or* of :class:`object` otherwise. This subclass exposes
    the following public instance variables:

    expr_alias_cls : EnumType
        Value of the ``enum_type`` parameter passed to this method (i.e., the
        enumeration constraining this expression). For duck typing between
        subclasses dynamically synthesized by this and the similar
        :func:`expr_alias` function, this variable shares the same name as that
        defined by those subclasses.

    Parameters
    ----------
    expr : str
        Arbitrarily complex Python expression suitable for use as both the
        left- and right-hand sides of Python assignment statements, typically
        evaluating to the value of an arbitrarily nested dictionary key:
        * Prefixed by the name of the instance variable to which this
          dictionary is bound in instances of the class containing this
          descriptor (e.g., ``self._config``).
        * Suffixed by one or more ``[``- and ``]``-delimited key lookups into
          the same dictionary (e.g.,
          ``['variable settings']['noise']['dynamic noise']``).
    enum_type : EnumType
        Enumeration constraining this expression. Specifically, the uppercased
        string value of this expression *must* always be the name of a member
        of this enumeration. If this is *not* the case, an exception is raised.
    base_classes : optional[SequenceTypes]
        Sequence of all base classes of the class dynamically synthesized by
        this function for the returned expression alias data descriptor,
        usually used to unify all such descriptors under a common abstract base
        class.  Defaults to the empty tuple, equivalent to the 1-tuple
        ``(object,)`` containing only the root base class of all classes.

    Returns
    ----------
    base_classes
        Enemuration-specific expression alias data descriptor as detailed
        above, guaranteed to be an instance of all passed base classes.

    See Also
    ----------
    :func:`expr_alias`
        Further details.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.text.string import strs
    from betse.util.type.cls import classes

    # Validate the names of all members of this enumeration to be uppercase.
    for enum_member in enum_type:
        if not strs.is_uppercase(enum_member.name):
            raise BetseEnumException(
                'Enumeration member name "{}" not uppercase.'.format(
                    enum_member.name))

    # Dictionary eventually containing only the following three keys:
    # "__init__", "__get__", and "__set__".
    expr_alias_class_methods = {
        # Type of all values accepted by this __set__() method.
        '__expr_enum_alias_type': enum_type,
    }

    # Snippet of Python expr declaring these special methods. See the
    # expr_alias() function for further details.
    expr_alias_class_method_defs = '''
def __init__(self_descriptor, __expr_enum_alias_type=__expr_enum_alias_type):
    self_descriptor.expr_alias_cls = __expr_enum_alias_type


# Convert this variable's low-level string into a high-level enumeration member.
def __get__(self_descriptor, self, cls):

    # If this data descriptor is accessed as a class attribute, return this data
    # descriptor as is. See enum_expr() for further details.
    if self is None:
        return self_descriptor

    # Lowercase name of the corresponding enumeration member to be returned.
    enum_member_name = {expr}

    # If this expression failed to evaluate to a string, raise an exception.
    if not isinstance(enum_member_name, str):
        raise BetseTypeException(
            'Expression alias value {{!r}} not a string.'.format(
                enum_member_name))

    # If this member name is *NOT* strictly lowercase, raise an exception.
    if not enum_member_name.islower():
        raise BetseTypeException(
            'Expression alias string {{!r}} not strictly lowercase.'.format(
                enum_member_name))

    # Uppercase member name.
    enum_member_name = enum_member_name.upper()

    # If this member is unrecognized, raise an exception. For efficiency and
    # readability, this test duplicates the enums.is_member_name() tester.
    if enum_member_name not in (
        self_descriptor.expr_alias_cls.__members__):
        raise BetseEnumException(
            'Expression alias string {{!r}} unrecognized (i.e., '
            'not the name of an enumeration member of {{!r}}).'.format(
                enum_member_name.lower(),
                self_descriptor.expr_alias_cls))

    # Return the enumeration member with this name.
    return self_descriptor.expr_alias_cls[enum_member_name]


# Convert the passed high-level enumeration member into a low-level string.
def __set__(self_descriptor, self, enum_member):
    # If the passed value is *NOT* a member of this enumeration, raise an
    # exception. For both efficiency and readability, this logic duplicates that
    # of the enums.is_member() function.
    if enum_member not in self_descriptor.expr_alias_cls:
        raise BetseEnumException(
            'Expression alias value {{!r}} unrecognized (i.e., not an '
            'enumeration member of {{!r}}).'.format(
                enum_member, self_descriptor.expr_alias_cls))

    # Set this expression to the lowercased name of this enumeration member.
    #from betse.util.io.log import logs
    #logs.log_info('Setting enumeration to: {{}}'.format(enum_member.name.lower()))
    {expr} = enum_member.name.lower()
'''.format(expr=expr)
    # print('expr_alias code:\n' + expr_alias_class_method_defs)

    # Define these methods and this dictionary containing these methods.
    exec(expr_alias_class_method_defs, globals(), expr_alias_class_methods)

    # Prevent input parameters passed into this snippet by this exec() call from
    # polluting the attribute namespace of this class.
    del expr_alias_class_methods['__expr_enum_alias_type']

    # Descriptor class with this name containing only these methods.
    expr_alias_class = classes.define_class(
        base_classes=base_classes,
        class_name=_get_expr_alias_class_name(),
        class_attr_name_to_value=expr_alias_class_methods,
    )

    # Instantiate and return the singleton descriptor for this class.
    return expr_alias_class()

# ....................{ PRIVATE                           }....................
def _get_expr_alias_class_name() -> str:
    '''
    Name of the class of the next data descriptor created and returned by the
    next call to an expression alias function (e.g., :func:`expr_alias`),
    guaranteed to be unique across all such classes.

    Since the human-readability of this name is neither required nor desired,
    this is a non-human-readable name efficiently constructed from a private
    prefix and a unique arbitrary identifier.
    '''

    # Global variable set below.
    global _EXPR_ALIAS_ID

    # Increment this identifier, preserving uniqueness.
    _EXPR_ALIAS_ID += 1

    # Return a unique name suffixed by this identifier.
    return '__ExprAlias' + str(_EXPR_ALIAS_ID)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Decorators skipping tests.

These decorators conditionally mark their decorated tests as skipped depending
on whether the conditions signified by the passed parameters are satisfied
(e.g., the importability of the passed module name).
'''

# ....................{ IMPORTS                            }....................
import pytest
from betse.util.type.types import (
    type_check,
    CallableTypes,
    ClassType,
    MappingOrNoneTypes,
    SequenceOrNoneTypes,
    StrOrNoneTypes,
)
from betse.util.test.pytest.mark import pytmark

# ....................{ IMPORTS ~ private                  }....................
# Sadly, the following imports require private modules and packages.
from _pytest.runner import Skipped

# ....................{ SKIP                               }....................
skip_if = pytest.mark.skipif
'''
Conditionally skip the decorated test or fixture with the passed human-readable
justification if the passed boolean is ``False``.

Parameters
----------
boolean : bool
    Boolean to be tested.
reason : str
    Human-readable message justifying the skipping of this test or fixture.
'''


@type_check
def skip(reason: str):
    '''
    Unconditionally skip the decorated test with the passed human-readable
    justification.

    This decorator is intended to be called both directly as a function *and*
    indirectly as a decorator, which differs from both:

    * :func:`pytest.skip`, intended to be called only directly as a function.
      Attempting to call this function indirectly as a decorator produces
      extraneous ignorable messages on standard output resembling
      ``"SKIP [1] betse_test/unit/test_import.py:66: could not import
      'xdist'"``, for unknown (and probably uninteresting) reasons.
    * :func:`pytest.mark.skip`, intended to be called only indirectly as a
      decorator. Attempting to call this decorator directly as a function
      reduces to a noop, for unknown (and probably uninteresting) reasons.

    Parameters
    ----------
    reason : str
        Human-readable message justifying the skipping of this test.
    '''

    return skip_if(True, reason=reason)

# ....................{ SKIP ~ command                     }....................
@type_check
def skip_unless_command(pathname: str):
    '''
    Skip the decorated test unless a command with the passed path exists.

    This is the case if this path is neither:

    * The basename of an executable file in the current ``${PATH}``.
    * The relative or absolute path of an executable file.

    This decorator is typically applied to tests requiring optional third-party
    external commands.

    Parameters
    ----------
    pathname : str
        Basename or absolute or relative path of the executable file to
        inspect.

    Returns
    ----------
    pytest.skipif
        Decorator describing these requirements if unmet *or* the identity
        decorator successfully reducing to a noop otherwise.
    '''

    # Defer heavyweight imports.
    from betse.util.os.command import cmds

    # If this command exists, reduce this decoration to a noop.
    if cmds.is_command(pathname):
        return pytmark.noop
    # Else, skip this test with a human-readable justification.
    else:
        return skip('Command "{}" not found.'.format(pathname))

# ....................{ SKIP ~ host                        }....................
def skip_if_ci_gitlab():
    '''
    Skip the decorated test if tests are currently running under a remote
    GitLab-CI continuous integration (CI) host.

    Returns
    ----------
    pytest.skipif
        Decorator describing these requirements if unmet *or* the identity
        decorator successfully reducing to a noop otherwise.
    '''

    # Defer heavyweight imports.
    from betse.util.test import tstci

    # Skip this test or fixture if tests are remotely running under GitLab-CI.
    return skip_if(tstci.is_ci_gitlab(), reason=(
        'Test or fixture currently incompatible with GitLab-CI.'))

# ....................{ SKIP ~ lib                        }....................
@type_check
def skip_unless_matplotlib_anim_writer(writer_name: str):
    '''
    Skip the decorated test unless the external command underlying the
    matplotlib animation writer with the passed name is in the current
    ``${PATH}`` (e.g., for the ``imagemagick`` writer, if the ``convert``
    command  is found).

    Parameters
    ----------
    writer_name : str
        Name of the matplotlib animation writer to be inspected.

    Returns
    ----------
    pytest.skipif
        Decorator describing these requirements if unmet *or* the identity
        decorator successfully reducing to a noop otherwise.
    '''

    # Defer heavyweight imports.
    from betse.lib.matplotlib.writer import mplvideo

    # If this command exists, reduce this decoration to a noop.
    if mplvideo.is_writer(writer_name):
        return pytmark.noop
    # Else, skip this test with a human-readable justification.
    else:
        return skip(
            'Matplotlib animation writer "{}" either '
            'not found or unrecognized.'.format(writer_name))

# ....................{ SKIP ~ lib : optional              }....................
@type_check
def skip_unless_lib_runtime_optional(*lib_names: str):
    '''
    Skip the decorated test if one or more of the optional runtime dependencies
    of this application with the passed :mod:`setuptools`-specific project
    names are **unsatisfiable** (i.e., unimportable *or* of unsatisfactory
    version).

    Parameters
    ----------
    lib_names : str
        Tuple of the names of all :mod:`setuptools`-specific projects
        identifying these dependencies (e.g., ``NetworkX``).

    Returns
    ----------
    pytest.skipif
        Decorator describing these requirements if unmet *or* the identity
        decorator reducing to a noop otherwise.
    '''

    # Defer heavyweight imports.
    from betse.exceptions import BetseLibException
    from betse.lib import libs

    # Skip this test if one or more such dependences are unsatisfiable.
    #
    # Note that the lower-level libs.die_unless_requirements_dict_keys()
    # function is intentionally called in lieu of the higher-level
    # libs.die_unless_runtime_optional() function. Why? Because the latter
    # assumes the application metadata singleton to have already been
    # initialized. Due to chicken-and-egg issues during test collection, this
    # initialization cannot be guaranteed prior to such test collection. Ergo,
    # the lower-level approach is preferred.
    return _skip_if_callable_raises_exception(
        exception_type=BetseLibException,
        func=libs.die_unless_runtime_optional,
        args=lib_names,
    )

# ....................{ SKIP ~ lib : requirement           }....................
@type_check
def skip_if_requirement(*requirements_str: str):
    '''
    Skip the decorated test if one or more of the dependencies identified by
    the passed :mod:`setuptools`-formatted requirement strings are
    **satisfiable** (i.e., importable *and* of satisfactory version).

    Parameters
    ----------
    requirements_str : str
        Tuple of all :mod:`setuptools`-formatted requirement strings
        identifying these dependencies (e.g., ``Numpy >= 1.8.0``).

    Returns
    ----------
    pytest.skipif
        Decorator describing these requirements if unmet *or* the identity
        decorator reducing to a noop otherwise.
    '''

    # Defer heavyweight imports.
    from betse.exceptions import BetseLibException
    from betse.lib.libs import die_unless_runtime_optional
    # from betse.lib.setuptools import setuptool
    from betse.util.type.text.string import strjoin

    # Human-readable message justifying the skipping of this test or fixture.
    reason = 'Test or fixture currently incompatible with {}.'.format(
        strjoin.join_as_conjunction_double_quoted(*requirements_str))

    # Skip this test if one or more such dependences are satisfiable.
    return _skip_unless_callable_raises_exception(
        exception_type=BetseLibException,
        func=die_unless_runtime_optional,
        # func=setuptool.die_unless_requirements_str,
        args=requirements_str,
        reason=reason,
    )


@type_check
def skip_unless_requirement(*requirements_str: str):
    '''
    Skip the decorated test if one or more of the dependencies identified by
    the passed :mod:`setuptools`-formatted requirement strings are
    **unsatisfiable** (i.e., unimportable *or* of unsatisfactory version).

    Parameters
    ----------
    requirements_str : str
        Tuple of all :mod:`setuptools`-formatted requirement strings
        identifying these dependencies (e.g., ``Numpy >= 1.8.0``).

    Returns
    ----------
    pytest.skipif
        Decorator describing these requirements if unmet *or* the identity
        decorator reducing to a noop otherwise.
    '''

    # Defer heavyweight imports.
    from betse.exceptions import BetseLibException
    from betse.lib.libs import die_unless_runtime_optional
    # from betse.lib.setuptools import setuptool

    # Skip this test if one or more such dependences are unsatisfiable.
    return _skip_if_callable_raises_exception(
        exception_type=BetseLibException,
        func=die_unless_runtime_optional,
        # func=setuptool.die_unless_requirements_str,
        args=requirements_str,
    )

# ....................{ SKIP ~ module                      }....................
@type_check
def skip_unless_module(module_name: str, minimum_version: str = None):
    '''
    Skip the decorated test if the module with the passed name is unimportable
    *or* importable but of a version less than the passed minimum version if
    non-``None``.

    Note that tests requiring optional third-party Python dependencies should
    call the higher-level :func:`skip_unless_lib_runtime_optional` decorator
    instead, which implicitly validates the versions of those dependencies
    rather than requiring those versions be explicitly passed.

    Caveats
    ----------
    The higher-level :func:`skip_unless_lib_runtime_optional` and
    :func:`skip_unless_requirement` decorators (which are implemented in terms
    of robust :mod:`setuptools` machinery) should typically be called in lieu
    of this lower-level decorator (which is implemented in terms of fragile
    :mod:`importlib` machinery).

    Parameters
    ----------
    module_name : str
        Fully-qualified name of the module to be tested for.
    minimum_version : Optional[str]
        Optional minimum version of this module as a dot-delimited string
        (e.g., ``0.4.0``) to be tested for if any *or* ``None`` otherwise, in
        which case any version is acceptable. Defaults to ``None``.

    Returns
    ----------
    pytest.skipif
        Decorator describing these requirements if unmet *or* the identity
        decorator reducing to a noop otherwise.

    See Also
    ----------
    :func:`skip_unless_lib_runtime_optional`
        Higher-level decorator skipping the decorated test if the passed
        optional runtime dependency is unsatisfiable.
    '''

    return _skip_if_callable_raises_exception(
        exception_type=Skipped,
        func=pytest.importorskip,
        args=(module_name, minimum_version),
    )

# ....................{ SKIP ~ os                          }....................
def skip_if_os_macos():
    '''
    Skip the decorated test if tests are currently running under an operating
    system running the Apple macOS kernel.

    Returns
    ----------
    pytest.skipif
        Decorator describing these requirements if unmet *or* the identity
        decorator successfully reducing to a noop otherwise.
    '''

    # Defer heavyweight imports.
    from betse.util.os.brand.macos import is_macos

    # Skip this test or fixture if tests are running under macOS.
    return skip_if(is_macos(), reason=(
        'Test or fixture currently incompatible with Apple macOS.'))


def skip_if_os_windows_vanilla():
    '''
    Skip the decorated test if tests are currently running under an operating
    system running the vanilla (i.e., neither Cygwin nor the Windows Subsystem
    for Linux (WSL)) Microsoft Windows kernel.

    Returns
    ----------
    pytest.skipif
        Decorator describing these requirements if unmet *or* the identity
        decorator successfully reducing to a noop otherwise.
    '''

    # Defer heavyweight imports.
    from betse.util.os.brand.windows import is_windows_vanilla

    # Skip this test or fixture if tests are running under Windows.
    return skip_if(is_windows_vanilla(), reason=(
        'Test or fixture currently incompatible with '
        'vanilla Microsoft Windows.'
    ))

# ....................{ SKIP ~ plugin                      }....................
def skip_unless_plugin_xdist():
    '''
    Skip the decorated test if the ``pytest-xdist`` plugin is *not* installed.

    This decorator is typically applied to tests requiring **process
    isolation** (i.e., isolating tests to dedicated subprocesses of the current
    test session).  While this plugin provides such isolation out-of-the-box,
    vanilla :mod:`pytest` does not. Hence, these tests *must* be skipped in the
    absence of this plugin.

    Examples of tests requiring process isolation include:

    * Unit tests testing importability. Since Python caches imports performed
      by the active Python interpreter (e.g., via :attr:`sys.modules`) *and*
      since the order in which :mod:`pytest` runs tests should be assumed to be
      non-deterministic, importability *cannot* be reliably tested within a
      single Python process.
    '''

    return skip_unless_module('xdist')

# ....................{ PRIVATE ~ skip                     }....................
@type_check
def _skip_if_callable_raises_exception(
    # Mandatory parameters.
    exception_type: ClassType,
    func: CallableTypes,

    # Optional parameters.
    args: SequenceOrNoneTypes = None,
    kwargs: MappingOrNoneTypes = None,
):
    '''
    Skip the decorated test if calling the passed callable with the passed
    positional and keyword arguments raises an exception of the passed type.

    Specifically, if calling this callable raises:

    * The passed type of exception, this test is marked as skipped.
    * Any other type of exception, this test is marked as a failure.
    * No exception, this test continues as expected.

    Parameters
    ----------
    exception_type : ClassType
        Type of exception expected to be raised by this callable.
    func : CallableTypes
        Callable to be called.
    args : SequenceOrNoneTypes
        Sequence of all positional arguments to unconditionally pass to the
        passed callable if any *or* ``None`` otherwise. Defaults to ``None``.
    kwargs : MappingOrNoneTypes
        Dictionary of all keyword arguments to unconditionally pass to the
        passed callable if any *or* ``None`` otherwise. Defaults to ``None``.

    Returns
    ----------
    pytest.skipif
        Decorator skipping this test if this callable raises this exception
        *or* the identity decorator reducing to a noop otherwise.
    '''

    # Default all unpassed arguments to sane values.
    if args is None:
        args = ()
    if kwargs is None:
        kwargs = {}

    # Attempt to call this callable with these arguments.
    try:
        func(*args, **kwargs)
    # If this callable raises an expected exception, skip this test.
    except exception_type as exception:
        return skip(str(exception))
    # If this callable raises an unexpected exception, fail this test. To
    # preserve this exception's stack trace, reraise this exception as is.
    except Exception:
        raise
        # # Print this exception's stacktrace to stderr.
        # stderrs.output_exception(heading=(
        #     'skip_unless_lib_runtime_optional{} '
        #     'raised unexpected exception:\n'.format(lib_names)))
        #
        # # Skip this test with this exception's message.
        # return skip(str(exception))
    # Else, this callable raised no exception. Silently reduce to a noop.
    else:
        return pytmark.noop


@type_check
def _skip_unless_callable_raises_exception(
    # Mandatory parameters.
    exception_type: ClassType,
    func: CallableTypes,

    # Optional parameters.
    args: SequenceOrNoneTypes = None,
    kwargs: MappingOrNoneTypes = None,
    reason: StrOrNoneTypes = None,
):
    '''
    Skip the decorated test unless calling the passed callable with the passed
    positional and keyword arguments raises an exception of the passed type.

    Specifically, if calling this callable raises:

    * The passed type of exception, this test continues as expected.
    * Any other type of exception, this test is marked as a failure.
    * No exception, this test is marked as skipped.

    Parameters
    ----------
    exception_type : ClassType
        Type of exception expected to be raised by this callable.
    func : CallableTypes
        Callable to be called.
    args : SequenceOrNoneTypes
        Sequence of all positional arguments to unconditionally pass to the
        passed callable if any *or* ``None`` otherwise. Defaults to ``None``.
    kwargs : MappingOrNoneTypes
        Dictionary of all keyword arguments to unconditionally pass to the
        passed callable if any *or* ``None`` otherwise. Defaults to ``None``.
    reason : StrOrNoneTypes
        Human-readable message justifying the skipping of this test or fixture.
        Defaults to ``None``, in which case this function utilizes a
        general-purpose message generically applicable to all tests.

    Returns
    ----------
    pytest.skipif
        Decorator skipping this test if this callable does *not* raise this
        exception *or* the identity decorator reducing to a noop otherwise.
    '''

    # Defer heavyweight imports.
    from betse.util.type.cls import classes

    # Default all unpassed arguments to sane values.
    if args is None:
        args = ()
    if kwargs is None:
        kwargs = {}
    if reason is None:
        reason = 'Test or fixture failed to raise exception "{}".'.format(
            classes.get_name_unqualified(exception_type))

    # Attempt to call this callable with these arguments.
    try:
        func(*args, **kwargs)
    # If this callable raises an expected exception, reduce to a noop.
    except exception_type:
        return pytmark.noop
    # If this callable raises an unexpected exception, fail this test. To
    # preserve this exception's stack trace, reraise this exception as is.
    except Exception:
        raise
    # Else, this callable raised no exception. Skip this test.
    else:
        return skip(reason)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
**Command runners** (i.e., functions running commands on behalf of callers).

Command Words Arguments
----------
Most runners accept a mandatory ``command_words`` parameter, a list of one or
more shell words comprising this command whose:

* Mandatory first item is either:

  * This command's absolute or relative path.
  * This command's basename, in which case the first command with that basename
    in the current ``${PATH}`` environment variable will be run. If no such
    command is found, an exception is raised.

* Optional subsequent items are this command's arguments (in order). Note that
  these arguments are passed as is to a low-level system call rather than
  intprereted by a high-level shell (e.g., ``/bin/sh`` on POSIX-compatible
  platforms) and hence should *not* be shell-quoted. Indeed, shell quoting
  these arguments is likely to result in erroneous command behaviour. The
  principal exception to this heuristic are **GNU-style long value options**
  (i.e., ``--``-prefixed options accepting a ``=``-delimited value), whose
  values should either be:

  * Passed as a separate argument *without* being shell-quoted.
  * Concatenated to the current ``--``-prefixed argument delimited by ``=`` and
    shell-quoted.

``Popen()`` Keyword Arguments
----------
Most runners accept the same optional keyword arguments accepted by the
:meth:`subprocess.Popen.__init__` constructor, including:

* ``cwd``, the absolute path of the current working directory (CWD) from which
  this command is to be run. Defaults to the current CWD. **Unfortunately, note
  that this keyword argument appears to be erroneously ignored on numerous
  platforms (e.g., Windows XP).** For safety, the
  :func:`betse.util.os.shell import shelldir.setting_cwd` context manager
  should typically be leveraged instead.
* ``timeout``, the maximum number of milliseconds this command is to be run
  for. Commands with execution time exceeding this timeout will be mercifully
  killed. Defaults to ``None``, in which case this command is run indefinitely.
'''

# ....................{ IMPORTS                           }....................
import subprocess
from betse.exceptions import BetseCommandException
from betse.util.io.log import logs
from betse.util.io.log.logenum import LogLevel
from betse.util.type.types import (
    type_check, MappingType, MappingOrNoneTypes, SequenceTypes)
from io import TextIOWrapper
from subprocess import CalledProcessError, Popen, PIPE, TimeoutExpired
from threading import Thread

# ....................{ GLOBALS                           }....................
BUFFER_SIZE_DEFAULT = -1
'''
Default subprocess buffer size for the current platform (synonymous with the
current :data:`io.DEFAULT_BUFFER_SIZE`) suitable for passing as the ``bufsize``
parameter accepted by :meth:`subprocess.Popen.__init__` method.
'''


BUFFER_SIZE_NONE = 0
'''
Unbuffered subprocess buffer size suitable for passing as the ``bufsize``
parameter accepted by :meth:`subprocess.Popen.__init__` method.

Both reading from and writing to an unbuffered subprocess is guaranteed to
perform exactly one system call (``read()`` and ``write()``, respectively) and
can return short (i.e., produce less bytes than the requested number of bytes).
'''


BUFFER_SIZE_LINE = 1
'''
Line-buffered subprocess buffer size suitable for passing as the ``bufsize``
parameter accepted by :meth:`subprocess.Popen.__init__` method.

Reading from a line-buffered subprocess is guaranteed to block until the
subprocess emits a newline, at which point all output emitted between that
newline inclusive and the prior newline exclusive is consumed.
'''

# ....................{ RUNNERS                           }....................
def run_or_die(
    command_words: SequenceTypes, popen_kwargs: MappingOrNoneTypes = None
) -> None:
    '''
    Run the passed command as a subprocess of the current Python process,
    raising an exception on subprocess failure *and* forwarding all
    standard output and error output by this subprocess to the standard output
    and error file handles of the current Python process.

    This exception contains the exit status of this subprocess.

    Parameters
    ----------
    command_words : SequenceTypes
        List of one or more shell words comprising this command.
    popen_kwargs : optional[MappingType]
        Dictionary of all keyword arguments to pass to the
        :meth:`subprocess.Popen.__init__` method. Defaults to ``None``, in
        which case the empty dictionary is assumed.

    Raises
    ----------
    CalledProcessError
        Exception raised on subprocess failure.
    '''

    # Avoid circular import dependencies.
    from betse.util.os.command import cmdexit

    # Run this command, raising an exception on command failure. For
    # reusability, reimplement the subprocess.check_call() function here rather
    # than explicitly call this function. The latter approach would require
    # duplicating logic between this and the get_exit_status() function.
    exit_status = get_exit_status(
        command_words=command_words, popen_kwargs=popen_kwargs)
    if cmdexit.is_failure(exit_status):
        raise CalledProcessError(exit_status, command_words)

# ....................{ GETTERS                           }....................
def get_exit_status(
    command_words: SequenceTypes, popen_kwargs: MappingOrNoneTypes = None
) -> int:
    '''
    Run the passed command as a subprocess of the current Python process,
    returning only the exit status of this subprocess *and* forwarding all
    standard output and error output by this subprocess to the standard output
    and error file handles of the current Python process.

    This function raises *no* exceptions on subprocess failure. To do so,
    consider calling the :func:`run` function instead.

    Parameters
    ----------
    command_words : SequenceTypes
        List of one or more shell words comprising this command.
    popen_kwargs : optional[MappingType]
        Dictionary of all keyword arguments to pass to the
        :meth:`subprocess.Popen.__init__` method. Defaults to ``None``, in
        which case the empty dictionary is assumed.

    Returns
    ----------
    int
        Exit status returned by this subprocess.
    '''

    # Avoid circular import dependencies.
    from betse.util.os.command.cmdexit import FAILURE_DEFAULT

    # Sanitize these arguments.
    popen_kwargs = _init_popen_kwargs(command_words, popen_kwargs)

    # Run this command *WITHOUT* raising an exception on command failure.
    try:
        exit_status = subprocess.call(command_words, **popen_kwargs)
    # If this command failed to halt before triggering a timeout, the "timeout"
    # keyword argument was passed *AND* this command has effectively failed.
    # Since the prior call has already guaranteeably terminated this command,
    # this exception is safely convertable into failure exit status.
    except TimeoutExpired:
        exit_status = FAILURE_DEFAULT

    # Return this exit status.
    return exit_status

# ....................{ GETTERS ~ stdout                  }....................
def get_stdout_or_die(
    command_words: SequenceTypes, popen_kwargs: MappingOrNoneTypes = None
) -> str:
    '''
    Run the passed command as a subprocess of the current Python process,
    raising an exception on subprocess failure *and* capturing and returning
    all stdout output by this subprocess.

    Parameters
    ----------
    command_words : SequenceTypes
        List of one or more shell words comprising this command.
    popen_kwargs : optional[MappingType]
        Dictionary of all keyword arguments to pass to the
        :meth:`subprocess.Popen.__init__` method. Defaults to ``None``, in
        which case the empty dictionary is assumed.

    Returns
    ----------
    str
        All stdout captured from this subprocess, stripped of all trailing
        newlines (as under most POSIX shells) *and* decoded with the current
        locale's preferred encoding (e.g., UTF-8).

    Raises
    ----------
    CalledProcessError
        Exception raised on subprocess failure.
    '''

    # Sanitize these arguments.
    popen_kwargs = _init_popen_kwargs(command_words, popen_kwargs)

    # Capture this command's stdout, raising an exception on command failure
    # (including failure due to an expired timeout).
    command_stdout = subprocess.check_output(command_words, **popen_kwargs)

    # Return this stdout, stripped of all trailing newlines.
    return command_stdout.rstrip('\n')


def get_output_interleaved_or_die(
    command_words: SequenceTypes, popen_kwargs: MappingOrNoneTypes = None
) -> str:
    '''
    Run the passed command as a subprocess of the current Python process,
    raising an exception on subprocess failure *and* capturing and returning
    all stdout and stderr output by this subprocess interleaved together (in
    arbitrary order).

    Parameters
    ----------
    command_words : SequenceTypes
        List of one or more shell words comprising this command.
    popen_kwargs : optional[MappingType]
        Dictionary of all keyword arguments to pass to the
        :meth:`subprocess.Popen.__init__` method. Defaults to ``None``, in
        which case the empty dictionary is assumed.

    Returns
    ----------
    str
        All stdout and stderr captured from this subprocess, interleaved
        together in output order, stripped of all trailing newlines (as under
        most POSIX shells) *and* decoded with the current locale's preferred
        encoding (e.g., UTF-8).

    Raises
    ----------
    CalledProcessError
        Exception raised on subprocess failure.
    '''

    # If these keyword arguments are empty, default to the empty dictionary.
    if not popen_kwargs:
        popen_kwargs = {}

    # Redirect stderr to stdout.
    popen_kwargs['stderr'] = subprocess.STDOUT

    # Capture and return this command's stdout and stderr.
    return get_stdout_or_die(
        command_words=command_words, popen_kwargs=popen_kwargs)

# ....................{ LOGGERS                           }....................
def log_output_or_die(
    # Mandatory arguments.
    command_words: SequenceTypes,

    # Optional arguments.
    stdout_log_level: LogLevel = LogLevel.INFO,
    stderr_log_level: LogLevel = LogLevel.ERROR,
    popen_kwargs: MappingOrNoneTypes = None,
) -> None:
    '''
    Run the passed command as a subprocess of the current Python process,
    raising an exception on subprocess failure *and* capturing and logging all
    stdout output by this subprocess with logging level ``INFO`` and all
    stderr output by this subprocess with logging level ``ERROR``.

    For space efficiency and logging responsiveness, this function captures and
    logs subprocess output in a line-buffered manner. Specifically:

    * Each line of stdout output by this subprocess is logged with logging
      level :attr:`LogLevel.INFO` with multi-threaded asynchronous I/O.
    * Each line of stdout output by this subprocess is logged with logging
      level :attr:`LogLevel.INFO` with multi-threaded asynchronous I/O.

    Parameters
    ----------
    command_words : SequenceTypes
        List of one or more shell words comprising this command.
    stdout_log_level : LogLevel
        Logging level with which all stdout output is logged. Defaults to
        :attr:`LogLevel.INFO`.
    stderr_log_level : LogLevel
        Logging level with which all stderr output is logged. Defaults to
        :attr:`LogLevel.ERROR`.
    popen_kwargs : optional[MappingType]
        Dictionary of all keyword arguments to pass to the
        :meth:`subprocess.Popen.__init__` method. Defaults to ``None``, in
        which case the empty dictionary is assumed.

    Raises
    ----------
    CalledProcessError
        Exception raised on subprocess failure.
    '''

    # Sanitize these arguments.
    popen_kwargs = _init_popen_kwargs(command_words, popen_kwargs)

    # Subprocess forked from this process, redirecting both stdout and stderr
    # to consumable pipes in a line-buffered manner.
    command_subprocess = Popen(
        args=command_words,
        stdout=PIPE,
        stderr=PIPE,
        bufsize=BUFFER_SIZE_LINE,
        **popen_kwargs
    )

    # Threads logging each line of stdout and stderr output by this subprocess
    # with the appropriate logging levels.
    stdout_logger = Thread(
        target=_log_pipe_lines,
        args=(command_subprocess.stdout, stdout_log_level))
    stderr_logger = Thread(
        target=_log_pipe_lines,
        args=(command_subprocess.stderr, stderr_log_level))

    # Run this subprocess in a multi-threaded manner logged by these
    # asynchronous threads, raising an exception on subprocess failure.
    stdout_logger.start()
    stderr_logger.start()


@type_check
def _log_pipe_lines(pipe: TextIOWrapper, log_level: LogLevel) -> None:
    '''
    Iteratively log each line buffered by the passed subprocess pipe as a new
    log message with the passed log level.

    Parameters
    ----------
    command_words : SequenceTypes
        List of one or more shell words comprising this command.
    popen_kwargs : Mapping
        Dictionary of keyword arguments to pass to the
        :meth:`subprocess.Popen.__init__` method.
    '''

    # Avoid circular import dependencies.
    from betse.util.io.iofiles import READLINE_EOF
    from betse.util.type.text.string import strs

    # With this pipe contextually opened for reading...
    with pipe:
        # For each plaintext line emitted by this pipe (falling back to the
        # standard EOF emitted by the readline() method in the event of
        # unexpected pipe closure)...
        for line in iter(pipe.readline, READLINE_EOF):
            # If this line signifies end-of-file (EOF), stop consuming lines.
            if line == READLINE_EOF:
                break

            # Message to be logged, stripped of trailing newline if any.
            message = strs.remove_newlines_suffix(line)

            # Log this line as a new log message with this logging level.
            logs.log_levelled(message, log_level)

# ....................{ PRIVATE                           }....................
@type_check
def _init_popen_kwargs(
    command_words: SequenceTypes, popen_kwargs: MappingOrNoneTypes
) -> MappingType:
    '''
    Sanitized dictionary of all keyword arguments to pass to the
    :class:`subprocess.Popen` callable when running the command specified by
    the passed shell words with the passed user-defined keyword arguments.

    `close_fds`
    ----------
    If the current platform is vanilla Windows *and* none of the ``stdin``,
    ``stdout``, ``stderr``, or ``close_fds`` arguments are passed, the latter
    argument will be explicitly set to ``False`` -- causing the command to be
    run to inherit all file handles (including stdin, stdout, and stderr) from
    the current process. By default, :class:`subprocess.Popen` documentation
    insists that:

    > On Windows, if ``close_fds`` is ``True`` then no handles will be
    > inherited by the child process.

    The child process will then open new file handles for stdin, stdout, and
    stderr. If the current terminal is a Windows Console, the underlying
    terminal devices and hence file handles will remain the same, in which case
    this is *not* an issue. If the current terminal is Cygwin-based (e.g.,,
    MinTTY), however, the underlying terminal devices and hence file handles
    will differ, in which case this behaviour prevents interaction between the
    current shell and the vanilla Windows command to be run below. In
    particular, all output from this command will be squelched.

    If at least one of stdin, stdout, or stderr are redirected to a blocking
    pipe, setting ``close_fds`` to ``False`` can induce deadlocks under certain
    edge-case scenarios. Since all such file handles default to ``None`` and
    hence are *not* redirected in this case, ``close_fds`` may be safely set to
    ``False``.

    On all other platforms, if ``close_fds`` is ``True``, no file handles
    *except* stdin, stdout, and stderr will be inherited by the child process.
    This function fundamentally differs in subtle (and only slightly documented
    ways) between vanilla Windows and all other platforms. These discrepancies
    appear to be harmful but probably unavoidable, given the philosophical gulf
    between vanilla Windows and all other platforms.

    Parameters
    ----------
    command_words : SequenceTypes
        List of one or more shell words comprising this command.
    popen_kwargs : optional[MappingType]
        Dictionary of all keyword arguments to be sanitized if any *or*
        ``None`` otherwise, in which case the empty dictionary is defaulted to.
    '''

    # Avoid circular import dependencies.
    from betse.util.os.command import cmds
    from betse.util.os.brand import windows
    from betse.util.os.shell import shellenv
    from betse.util.type.iterable.mapping import maptest

    # If this list of shell words is empty, raise an exception.
    if not command_words:
        raise BetseCommandException('Non-empty command expected.')

    # If these keyword arguments are empty, default to the empty dictionary.
    if popen_kwargs is None:
        popen_kwargs = {}

    # If the first shell word is this list is unrunnable, raise an exception.
    cmds.die_unless_command(command_words[0])

    # Log the command to be run before doing so.
    logs.log_debug('Running command: %s', ' '.join(command_words))

    # If this is vanilla Windows, sanitize the "close_fds" argument.
    if windows.is_windows_vanilla() and not maptest.has_keys(
        mapping=popen_kwargs,
        keys=('stdin', 'stdout', 'stderr', 'close_fds',)):
        popen_kwargs['close_fds'] = False

    # Isolate the current set of environment variables to this command,
    # preventing concurrent changes in these variables in the current process
    # from affecting this command's subprocess.
    popen_kwargs['env'] = shellenv.get_env()

    # Decode command output with the current locale's preferred encoding.
    popen_kwargs['universal_newlines'] = True

    # Return these keyword arguments.
    return popen_kwargs



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Unit tests exercising functionality defined by the
:mod:`betse.util.type.iterable.mapping.mapmerge` submodule.
'''

# ....................{ IMPORTS                           }....................
import pytest

# ....................{ TESTS                             }....................
def test_merge_maps() -> None:
    '''
    Unit test the :func:`betse.util.type.iterable.mapping.mapmerge.merge_maps`
    function.
    '''

    # Defer heavyweight imports.
    from betse.exceptions import BetseMappingException
    from betse.util.type.iterable.mapping import mapmerge
    from betse.util.type.iterable.mapping.mapmerge import MergeCollisionPolicy

    # Iterable of input dictionaries to be merged.
    hypnos = {
        'mercy':   'May the merciful gods, if indeed there be such,',
        'will':    'guard those hours when no power of the will,',
        'cunning': 'or drug that the cunning of man devises,',
        'sleep':   'can keep me from the chasm of sleep.',
    }
    the_call_of_cthulhu = {
        'mercy':     'The most merciful thing in the world, I think,',
        'inability': 'is the inability of the human mind',
        'correlate': 'to correlate all its contents.',
        'ignorance': 'We live on a placid island of ignorance',
        'infinity':  'in the midst of black seas of the infinity,',
        'voyage':    'and it was not meant that we should voyage far.',
    }
    the_colour_out_of_space = {
        'wild':  'West of Arkham the hills rise wild,',
        'woods': 'and there are valleys with deep woods',
        'cut':   'that no axe has ever cut.',

        # Key-value pair duplicated from the "hypnos" map to properly exercise
        # the "RAISE_EXCEPTION" merge policy.
        'mercy': 'May the merciful gods, if indeed there be such,',
    }
    the_other_gods = {
        'infinity': 'The vengeance of the infinite abysses...',
        'curse':    'That cursed, that damnable pit...',
        'mercy':    'Merciful gods of earth, I am falling into the sky!',
    }

    # Iterable of dictionaries containing one or more key collisions.
    mappings_nonunique = (hypnos, the_call_of_cthulhu, the_other_gods,)

    # Iterable of dictionaries containing *NO* key collisions.
    mappings_unique = (hypnos, the_colour_out_of_space)

    # Test whether merging dictionaries containing one or more key collisions
    # with exceptions raised on collisions raises the expected exception.
    with pytest.raises(BetseMappingException):
        mapmerge.merge_maps(
            mappings=mappings_nonunique,
            collision_policy=MergeCollisionPolicy.RAISE_EXCEPTION,
            is_values_copied=False,
        )

    # Test whether merging dictionaries containing no key collisions with
    # exceptions raised on collisions merges as expected.
    assert mapmerge.merge_maps(
        mappings=mappings_unique,
        collision_policy=MergeCollisionPolicy.RAISE_EXCEPTION,
        is_values_copied=True,
    ) == {
        'mercy':   'May the merciful gods, if indeed there be such,',
        'will':    'guard those hours when no power of the will,',
        'cunning': 'or drug that the cunning of man devises,',
        'sleep':   'can keep me from the chasm of sleep.',
        'wild':    'West of Arkham the hills rise wild,',
        'woods':   'and there are valleys with deep woods',
        'cut':     'that no axe has ever cut.',
    }

    # Test whether merging dictionaries containing key collisions with
    # precedence given to prior dictionaries succeeds as expected.
    assert mapmerge.merge_maps(
        mappings=mappings_nonunique,
        collision_policy=MergeCollisionPolicy.PREFER_FIRST,
        is_values_copied=False,
    ) == {
        'mercy':     'May the merciful gods, if indeed there be such,',
        'will':      'guard those hours when no power of the will,',
        'cunning':   'or drug that the cunning of man devises,',
        'sleep':     'can keep me from the chasm of sleep.',
        'inability': 'is the inability of the human mind',
        'correlate': 'to correlate all its contents.',
        'ignorance': 'We live on a placid island of ignorance',
        'infinity':  'in the midst of black seas of the infinity,',
        'voyage':    'and it was not meant that we should voyage far.',
        'curse':     'That cursed, that damnable pit...',
    }

    # Test whether merging dictionaries containing key collisions with
    # precedence given to subsequent dictionaries succeeds as expected.
    assert mapmerge.merge_maps(
        mappings=mappings_nonunique,
        collision_policy=MergeCollisionPolicy.PREFER_LAST,
        is_values_copied=True,
    ) == {
        'infinity':  'The vengeance of the infinite abysses...',
        'curse':     'That cursed, that damnable pit...',
        'mercy':     'Merciful gods of earth, I am falling into the sky!',
        'inability': 'is the inability of the human mind',
        'correlate': 'to correlate all its contents.',
        'ignorance': 'We live on a placid island of ignorance',
        'voyage':    'and it was not meant that we should voyage far.',
        'will':      'guard those hours when no power of the will,',
        'cunning':   'or drug that the cunning of man devises,',
        'sleep':     'can keep me from the chasm of sleep.',
    }



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **callable** (e.g., function, lambda, method, property) facilities.
'''

# ....................{ IMPORTS                           }....................
# from betse.util.io.log import logs
from betse.util.type.types import (
    type_check,
    CallableTypes,
    CallablePartialType,
    BuiltinFunctionType,
    BuiltinMethodType,
    FunctionType,
    FunctionTypes,
    GeneratorType,
    MethodType,
    MethodTypes,
    MappingOrNoneTypes,
    SequenceOrNoneTypes,
    StrOrNoneTypes,
)
from functools import partial, wraps

if False: wraps  # silence contemptible IDE warnings

# ....................{ CONSTANTS                         }....................
TYPE_TO_NAME = {
    BuiltinFunctionType: 'builtin',
    BuiltinMethodType:   'method',
    FunctionType:        'function',
    GeneratorType:       'generator',
    MethodType:          'method',
}
'''
Dictionary mapping from the class of each possible callable to the lowercase
human-readable type of that class, ignoring low-level implementation details
(e.g., the distinction between built-in and user-defined methods).
'''

# ....................{ TESTERS                           }....................
def is_function(obj) -> bool:
    '''
    ``True`` only if the passed object is a function.
    '''

    return isinstance(obj, FunctionTypes)


def is_method(obj) -> bool:
    '''
    ``True`` only if the passed object is a **method** (i.e., either a built-in
    or user-defined function bound at instantiation time to a specific object).
    '''

    return isinstance(obj, MethodTypes)

# ....................{ GETTERS                           }....................
@type_check
def get_name(func: CallableTypes) -> str:
    '''
    Machine-readable name of the passed callable.

    Parameters
    ----------
    func: CallableTypes
        Callable to get the name of.

    Returns
    ----------
    str
        Machine-readable name of this callable.

    See Also
    ----------
    :func:`to_str`
        Function getting a human-readable name of this callable.
    '''

    # The easy things in life are free.
    return func.__name__


@type_check
def get_doc_or_none(func: CallableTypes) -> StrOrNoneTypes:
    '''
    Human-readable docstring documenting the passed callable if any *or*
    ``None`` otherwise.

    Parameters
    ----------
    func: CallableTypes
        Callable to get the docstring of.

    Returns
    ----------
    StrOrNoneTypes
        Either:
        * If this callable has a docstring, this docstring.
        * Else, ``None``.
    '''

    return getattr(func, '__doc__', None)

# ....................{ SETTERS                           }....................
@type_check
def set_doc(func: CallableTypes, doc: str) -> None:
    '''
    Set the human-readable docstring documenting the passed callable to the
    passed string.

    Parameters
    ----------
    func: CallableTypes
        Callable to set the docstring of.
    doc : str
        Docstring to set on this callable.
    '''

    func.__doc__ = doc

# ....................{ CONVERTERS                        }....................
@type_check
def to_str(func: CallableTypes) -> str:
    '''
    Human-readable string describing the type and name of the passed callable
    (e.g., ``method Pala()``).

    For generality, this string's:

    * First character is guaranteed to be lowercase.
    * Last character is guaranteed to be the ``)`` delimiter.

    Parameters
    ----------
    func: CallableTypes
        Callable to describe.

    Returns
    ----------
    str
        Human-readable string describing this callable.
    '''

    # Type of this callable, defaulting to merely "callable" if unrecognized.
    func_type = TYPE_TO_NAME.get(type(func), 'callable')
    # print('func: {}, type: {}'.format(func.__name__, type(func)))

    # Return a human-readable string describing this callable.
    return '{} {}()'.format(func_type, func.__name__)


@type_check
def to_str_capitalized(func: CallableTypes) -> str:
    '''
    Human-readable string describing the type and name of the passed callable
    whose first character is capitalized (e.g., ``Method pala()``).

    See Also
    ----------
    :func:`to_str`
        Further details.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.text.string import strs

    # Return a capitalized human-readable string describing this callable.
    return strs.uppercase_char_first(to_str(func))

# ....................{ MAKERS                            }....................
@type_check
def make_partial(
    # Mandatory parameters.
    func: CallableTypes,

    # Optional parameters.
    args: SequenceOrNoneTypes = None,
    kwargs: MappingOrNoneTypes = None,
    doc: StrOrNoneTypes = None,
) -> CallablePartialType:
    '''
    **Partial callable** (i.e., callable whose signature is reduced by this
    function call to unconditionally pass one or more positional or keyword
    arguments to the passed callable) calling the passed callable, statically
    passed the passed positional and keyword arguments, and documented by the
    passed docstring.

    Parameters
    ----------
    func: CallableTypes
        Callable to partialize.
    args : SequenceOrNoneTypes
        Sequence of all positional arguments to unconditionally pass to the
        passed callable if any *or* ``None`` otherwise. Defaults to ``None``.
    kwargs : MappingOrNoneTypes
        Dictionary of all keyword arguments to unconditionally pass to the
        passed callable if any *or* ``None`` otherwise. Defaults to ``None``.
    doc : StrOrNoneTypes
        Docstring to document the returned callable with if any *or* ``None``
        otherwise. If ``None`` *and* the passed callable is documented, the
        returned callable will share the same docstring. Defaults to ``None``.

    Returns
    ----------
    CallablePartialType
        Partial callable synthesized from the passed callable and arguments.
    '''

    # Default all unpassed arguments to sane values.
    if args is None:
        args = ()
    if kwargs is None:
        kwargs = {}
    if doc is None:
        doc = get_doc_or_none(func)

    # Partial callable synthesized from the passed callable and arguments.
    func_partial = partial(func, *args, **kwargs)

    # If documenting this callable, do so.
    if doc is not None:
        set_doc(func=func_partial, doc=doc)

    # Return this partial callable.
    return func_partial



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Class hierarchy designing **simulation phase requirements** (i.e., objects
encapsulating the current state of a single simulation feature for a given
simulation phase).
'''

# ....................{ IMPORTS                            }....................
from abc import ABCMeta
from betse.science.phase.phasecls import SimPhase
# from betse.util.io.log import logs
from betse.util.type.decorator.deccls import abstractproperty
from betse.util.type.types import (
    type_check,
    CallableTypes,
    EnumMemberType,
    MappingOrNoneTypes,
)

# ....................{ SUPERCLASSES                       }....................
class SimPhaseRequirementABC(metaclass=ABCMeta):
    '''
    Abstract base class of all **simulation phase requirement** (i.e., object
    encapsulating the current state of a single simulation feature for a given
    simulation phase) subclasses.
    '''

    # ..................{ PROPERTIES ~ abstract              }..................
    # Subclasses are required to implement the following abstract properties.

    @abstractproperty
    def name(self) -> str:
        '''
        Human-readable name of this requirement (e.g., "extracellular spaces").

        For usability, the first character of this name should typically be
        lower- rather than uppercase.
        '''

        pass


    @abstractproperty
    def is_satisfied(self) -> CallableTypes:
        '''
        Callable (e.g., function, lambda) passed only the current simulation
        phase, returning ``True`` only if this requirement is enabled in this
        phase.
        '''

        pass


    @abstractproperty
    def set_satisfied(self) -> CallableTypes:
        '''
        Callable (e.g., function, lambda) passed only the current simulation
        phase, enabling this requirement in this phase.
        '''

        pass

# ....................{ SUBCLASSES ~ requirement           }....................
class SimPhaseRequirement(SimPhaseRequirementABC):
    '''
    **Simulation phase requirement** (i.e., high-level object encapsulating the
    current state of a single simulation feature, such as extracellular
    spaces).

    Caveats
    ----------
    The higher-level :class:`SimPhaseRequirements` class, which permits
    multiple instances of this lower-level class to be efficiently composed
    together via an immutable set-like API, is strongly preferable for external
    usage. Instances of this lower-level class should typically *only* be
    instantiated by the companion :mod:`betse.science.phase.require.phasereqs`
    submodule.

    To preserve hashability and hence usability in dictionaries and sets, this
    requirement is immutable. Callers should avoid breaking encapsulation to
    modify this requirement.

    Attributes
    ----------
    is_satisfied : CallableTypes
        Callable (e.g., function, lambda) passed the current simulation phase,
        returning ``True`` only if this requirement is satisfied by this phase.
    set_satisfied : CallableTypes
        Callable (e.g., function, lambda) passed the current simulation phase,
        modifying this phase as needed to ensure this requirement is satisfied
        by this phase.
    name : str
        Human-readable name of this requirement (e.g., ``extracellular
        spaces``). For simplicity, the first character of this name should
        typically be lower- rather than uppercase.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(
        self,
        name: str,
        is_satisfied: CallableTypes,
        set_satisfied: CallableTypes,
    ) -> None:
        '''
        Initialize this requirement.

        Parameters
        ----------
        _name : str
            Human-readable name of this requirement (e.g., ``extracellular
            spaces``). For simplicity, the first character of this name should
            typically be lower- rather than uppercase.
        _is_satisfied : CallableTypes
            Callable (e.g., function, lambda) passed only the current
            simulation phase, returning ``True`` only if this requirement is
            enabled in this phase.
        _set_satisfied : CallableTypes
            Callable (e.g., function, lambda) passed only the current
            simulation phase, enabling this requirement in this phase.
        '''

        # Initialize our superclass.
        super().__init__()

        # Classify all passed parameters.
        self._is_satisfied = is_satisfied
        self._set_satisfied = set_satisfied
        self._name = name

    # ..................{ PROPERTIES                         }..................
    # Subclasses are required to implement the following abstract properties.

    @property
    def name(self) -> str:
        return self._name

    @property
    def is_satisfied(self) -> CallableTypes:
        return self._is_satisfied

    @property
    def set_satisfied(self) -> CallableTypes:
        return self._set_satisfied

# ....................{ SUBCLASSES ~ requirement : body    }....................
class SimPhaseRequirementEmbodied(SimPhaseRequirement):
    '''
    String-based simulation phase requirement, initialized by strings defining
    this requirement's functions rather than actual functions.

    This requirement is a caller convenience simplifying initialization in the
    common case of a requirement whose functions are definable as strings.

    Attributes
    ----------
    _func_bodies : str
        String of Python code with which the :attr:`is_satisfied` and
        :attr:`set_satisfied` functions are defined. Technically, this string
        need *not* be classified as an instance variable. Pragmatically, doing
        so assists debugging elsewhere in the codebase.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(
        self,

        # Mandatory parameters.
        is_satisfied_body: str,
        set_satisfied_body: str,

        # Optional parameters.
        body_attrs: MappingOrNoneTypes = None,

        # Parental parameters.
        *args, **kwargs
    ) -> None:
        '''
        Initialize this requirement.

        Parameters
        ----------
        is_satisfied_body: str
            String of zero or more arbitrary complex Python statements
            comprising the body of a dynamically defined function:

            * Passed only the current simulation phase.
            * Returning ``True`` only if this requirement is enabled in this
              phase.
        set_satisfied_body: str
            String of zero or more arbitrary complex Python statements
            comprising the body of a dynamically defined function:

            * Passed only the current simulation phase.
            * Enabling this requirement in this phase.
        body_attrs : MappingOrNoneTypes
            Dictionary mapping from the name to value of each attribute
            globally exposed to the bodies of these dynamically defined
            functions. Defaults to ``None``, in which case these bodies may
            only reference:

            * The ``phase`` attribute defining the current simulation phase.
            * The ``SimPhase`` class.
            * the ``type_check`` decorator.

        All remaining parameters are passed as is to the superclass
        :meth:`SimPhaseRequirement.__init__` method.
        '''

        # In unpassed, default this parameter to the empty dictionary.
        if body_attrs is None:
            body_attrs = {}

        # Dictionary mapping from the name to value of each attribute globally
        # exposed to the declaration of these dynamically defined functions.
        #
        # Since the exec() statement called below adds key-value pairs
        # providing these functions to this dictionary, this dictionary is
        # copied from the passed dictionary to avoid mutating caller data.
        func_globals = {
            'type_check': type_check,
            'SimPhase': SimPhase,
        }
        func_globals.update(body_attrs)

        # Dictionary mapping from the name to value of each such function.
        funcs = {}

        # Raw string defining the functions to be passed to our superclass,
        # classified to assist debugging elsewhere in the codebase.
        self._func_bodies = '''
@type_check
def is_satisfied(phase: SimPhase) -> bool:
    {is_satisfied_body}

@type_check
def set_satisfied(phase: SimPhase) -> None:
    {set_satisfied_body}
'''.format(
    is_satisfied_body=is_satisfied_body,
    set_satisfied_body=set_satisfied_body,
)

        # Dynamically define these functions.
        exec(self._func_bodies, func_globals, funcs)
        # logs.log_debug('requirement: %s; funcs: %r; functions: %s',
        #     kwargs['name'], funcs, func_bodies,)

        # Initialize our superclass with these functions.
        super().__init__(
            *args,
            is_satisfied=funcs['is_satisfied'],
            set_satisfied=funcs['set_satisfied'],
            **kwargs)


class SimPhaseRequirementBoolExpr(SimPhaseRequirementEmbodied):
    '''
    String-based boolean simulation phase requirement, initialized by a single
    **boolean expression string** (i.e., string dynamically evaluating to a
    single gettable and settable boolean value).

    This requirement is a caller convenience simplifying initialization in the
    common case of a requirement reducing to a single boolean.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(self, bool_expr: str, *args, **kwargs) -> None:
        '''
        Initialize this requirement.

        Parameters
        ----------
        bool_expr : str
            Arbitrarily complex Python expression relative to the current
            simulation phase evaluating to a gettable and settable attribute
            whose value is this requirement's boolean state (e.g.,
            ``phase.p.is_ecm``). This expression may (and typically should)
            refer to the ``phase`` variable, bound to the current simulation
            phase.

        All remaining parameters are passed as is to the superclass
        :meth:`SimPhaseRequirementEmbodied.__init__` method.
        '''

        # Initialize our superclass with all passed parameters.
        super().__init__(
            *args,
            is_satisfied_body='return ' + bool_expr,
            set_satisfied_body=bool_expr + ' = True',
            **kwargs)


class SimPhaseRequirementEnumExpr(SimPhaseRequirementEmbodied):
    '''
    String-based enumeration member simulation phase requirement, initialized
    by a single **enumeration member expression string** (i.e., string
    dynamically evaluating to a single gettable and settable enumeration member
    value).

    This requirement is a caller convenience simplifying initialization in the
    common case of a requirement reducing to a single enumeration member.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(
        self,
        enum_expr: str,
        enum_member: EnumMemberType,
        *args, **kwargs
    ) -> None:
        '''
        Initialize this requirement.

        Parameters
        ----------
        enum_expr : str
            Arbitrarily complex Python expression relative to the current
            simulation phase evaluating to a gettable and settable attribute
            whose value is this requirement's enumeration state (e.g.,
            ``phase.p.solver_type``). This expression may (and typically
            should) refer to the ``phase`` variable, bound to the current
            simulation phase.
        enum_member : EnumMemberType
            Member of this enumeration satisfying this requirement (e.g.,
            ``SolverType.FULL``).

        All remaining parameters are passed as is to the superclass
        :meth:`SimPhaseRequirementEmbodied.__init__` method.
        '''

        # Bodies of the functions dynamically defined by our superclass.
        is_satisfied_body = 'return {} is {}'.format(
            enum_expr, enum_member.name)
        set_satisfied_body = '{} = {}'.format(
            enum_expr, enum_member.name)

        # Dictionary mapping from the name to value of this enumeration
        # member, locally exposing this member to the above bodies.
        body_attrs = {enum_member.name: enum_member}

        # Initialize our superclass with all passed parameters.
        super().__init__(
            *args,
            is_satisfied_body=is_satisfied_body,
            set_satisfied_body=set_satisfied_body,
            body_attrs=body_attrs,
            **kwargs)



#!/usr/bin/env python3
# ....................{ LICENSE                            }....................
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

# ....................{ IMPORTS                            }....................
import numpy as np
from betse.science import sim_toolbox as stb
from betse.util.io.log import logs
from scipy.interpolate import SmoothBivariateSpline

# ....................{ FUNCTIONS                          }....................
def getDeformation(sim, cells, t, p):
    """
    Calculate the deformation of the cell cluster under the action of
    intracellular forces and pressures, assuming steady-state (slow) changes.

    The method assumes that material is incompressible and total volume is
    conserved.

    The "galvanotropism" mechanism assumes growing and that ends of microtubules
    exert a cell-deforming force. Microtubules are in turn influenced by the
    electric field.

    If studying hydrostatic pressure deformations under osmotic influx, first,
    the equation of linear elastic motion is used to calculate deformation
    assuming full compressibility.

    The divergence of the resulting deformation field is calculated. An internal
    reaction pressure is calculated from the divergence. The gradient of the
    reaction pressure is subtracted from the initial solution to create a
    divergence-free (volume conserved) deformation field.
    """

    # Determine action forces
    #---------------------------------------------------------------------------------

    # calculate the gradient of any applied pressures:
    gPP = (sim.P_cells[cells.cell_nn_i[:, 1]] - sim.P_cells[cells.cell_nn_i[:, 0]]) / (cells.nn_len)

    gPx = -gPP*cells.nn_tx
    gPy = -gPP*cells.nn_ty

    sim.gPxc = np.dot(cells.M_sum_mems, gPx) / cells.num_mems
    sim.gPyc = np.dot(cells.M_sum_mems, gPy) / cells.num_mems

    # deformation by "galvanotropic" mechanism (electrostrictive forces
    # influenced by biology, e.g. cytoskeletal).
    # Fx = (1 / p.lame_mu) * (sim.gPxc + sim.E_cell_x*p.galvanotropism)
    # Fy = (1 / p.lame_mu) * (sim.gPyc + sim.E_cell_y*p.galvanotropism)
    Fx = (1 / p.lame_mu) * (sim.gPxc)
    Fy = (1 / p.lame_mu) * (sim.gPyc)

    # Calculate flow under body forces using time-independent linear elasticity
    # equation.
    dxo = np.dot(cells.lapGJinv, -Fx)
    dyo = np.dot(cells.lapGJinv, -Fy)

    # Deformation must be made divergence-free. To do so, use the
    # Helmholtz-Hodge decomposition method.
    _, d_cells_x, d_cells_y, _, _, _ = cells.HH_cells(
        dxo, dyo, rot_only=True, bounds_closed=p.fixed_cluster_bound)

    if p.deform_osmo:
        # Calculate the pressure gradient resulting from finite-divergence
        # osmotic-pressure induced water flows to cells.
        gPP = (sim.PP[cells.cell_nn_i[:, 1]] - sim.PP[cells.cell_nn_i[:, 0]]) / (cells.nn_len)

        dx = -gPP * cells.nn_tx
        dy = -gPP * cells.nn_ty

        dxco = np.dot(cells.M_sum_mems, dx) / cells.num_mems
        dyco = np.dot(cells.M_sum_mems, dy) / cells.num_mems

        # _, dxc, dyc, _, _, _ = cells.HH_cells(dxco, dyco, rot_only=True,
        #                                                           bounds_closed=p.fixed_cluster_bound)

        d_cells_x = d_cells_x + dxco
        d_cells_y = d_cells_y + dyco

    sim.d_cells_x = d_cells_x + sim.E_cell_x*p.galvanotropism
    sim.d_cells_y = d_cells_y + sim.E_cell_y*p.galvanotropism

    # sim.d_cells_x = d_cells_x*1
    # sim.d_cells_y = d_cells_y*1


def timeDeform(sim, cells, t, p):
    """
    Calculates the deformation of the cell cluster under the action
    of intracellular pressure, considering the full time-dependent
    linear elasticity equation for an incompressible medium.

    The solution method for this equation is similar to the
    steady-state method of deformation(). First the displacement
    field is calculated assuming compressibility,
    a reaction pressure is calculated from the divergence of the
    initial field, and the gradient of the internal pressure is
    subtracted from the initial field to produce a divergence
    free solution.

    This method is working much better than the timeDeform_o()
    so is presently in active use.

    """

    # # Check for the adequacy of the time step:
    # step_check = (p.dt / (2 * p.cell_radius)) * np.sqrt(p.lame_mu / 1000)
    #
    # if step_check > 1.0:
    #     new_ts = (0.9 * 2 * p.cell_radius) / (np.sqrt(p.lame_mu / 1000))
    #
    #     raise BetseSimException(
    #         'Time dependent deformation is tricky business, requiring a small time step! '
    #         'The time step you are using is too large to bother going further with. '
    #         'Please set your time step to ' + str(new_ts) + ' and try again.')
    #
    k_const = (p.dt ** 2) * (p.lame_mu / 1000)

    # # Determine action forces ------------------------------------------------

    # pressures:
    if p.deform_osmo is True:

        Pcell = sim.P_cells +  sim.PP

    else:

        Pcell = sim.P_cells

    # calculate the gradient of any applied pressures:
    gPP = (Pcell[cells.cell_nn_i[:, 1]] - Pcell[cells.cell_nn_i[:, 0]]) / (cells.nn_len)

    gPx = -gPP * cells.nn_tx
    gPy = -gPP * cells.nn_ty

    sim.gPxc = np.dot(cells.M_sum_mems, gPx) / cells.num_mems
    sim.gPyc = np.dot(cells.M_sum_mems, gPy) / cells.num_mems

    # deformation by "galvanotropic" mechanism (electrostrictive forces influenced by biology, e.g. cytoskeletal):
    F_cell_x = (1 / p.lame_mu) * ( (1/sim.sigma) * sim.J_cell_x * sim.rho_cells * p.galvanotropism + sim.gPxc)
    F_cell_y = (1 / p.lame_mu) * ( (1/sim.sigma) * sim.J_cell_y * sim.rho_cells * p.galvanotropism + sim.gPyc)

    # -------------------------------------------------------------------------------------------------

    sim.dx_time.append(np.copy(sim.d_cells_x))  # append the solution to the time-save vector
    sim.dy_time.append(np.copy(sim.d_cells_y))

    # Initial value solution--------------------------------------------------------------------------------
    if t == 0.0:

        wave_speed = np.sqrt(p.lame_mu / 1000)
        wave_speed = np.float(wave_speed)
        wave_speed = np.round(wave_speed, 2)

        logs.log_info(
            'Your wave speed is approximately: ' +
            str(wave_speed) + ' m/s '
        )

        logs.log_info('Try a world size of at least: ' + str(round((5 / 3) * (wave_speed / 500) * 1e6))
                      + ' um for resonance.')

        sim.d_cells_x = k_const * np.dot(cells.lapGJ, sim.dx_time[-1]) + (k_const / p.lame_mu) * F_cell_x + \
                        sim.dx_time[-1]
        sim.d_cells_y = k_const * np.dot(cells.lapGJ, sim.dy_time[-1]) + (k_const / p.lame_mu) * F_cell_y + \
                        sim.dy_time[-1]

    elif t > 0.0:

        # do the non-initial value, standard solution iteration:

        # calculate the velocity for viscous damping:
        d_ux_dt = (sim.dx_time[-1] - sim.dx_time[-2]) / (p.dt)
        d_uy_dt = (sim.dy_time[-1] - sim.dy_time[-2]) / (p.dt)

        gamma = ((p.dt ** 2) * (p.mu_tissue * p.lame_mu)) / (1000 * (2 * p.cell_radius))

        sim.d_cells_x = k_const * np.dot(cells.lapGJ, sim.dx_time[-1]) - gamma * d_ux_dt + \
                         (k_const / p.lame_mu) * F_cell_x + 2 * sim.dx_time[-1] - sim.dx_time[-2]

        sim.d_cells_y = k_const * np.dot(cells.lapGJ, sim.dy_time[-1]) - gamma * d_uy_dt + \
                         (k_const / p.lame_mu) * F_cell_y + 2 * sim.dy_time[-1] - sim.dy_time[-2]


    # Flow must be made divergence-free: use the Helmholtz-Hodge decomposition method:
    _, sim.d_cells_x, sim.d_cells_y, _, _, _ = cells.HH_cells(sim.d_cells_x, sim.d_cells_y, rot_only=True,
                                                              bounds_closed=p.fixed_cluster_bound)



    # calculate divergence of u  -----------------------------------------------------------------------
    # FIXME this might be a great place for HH decomposition

    # # first interpolate displacement field at membrane midpoints:
    # ux_mem = interp.griddata((cells.cell_centres[:, 0], cells.cell_centres[:, 1]), sim.d_cells_x,
    #     (cells.mem_mids_flat[:, 0], cells.mem_mids_flat[:, 1]), fill_value=0)
    #
    # uy_mem = interp.griddata((cells.cell_centres[:, 0], cells.cell_centres[:, 1]), sim.d_cells_y,
    #     (cells.mem_mids_flat[:, 0], cells.mem_mids_flat[:, 1]), fill_value=0)
    #
    # # get the component of the displacement field normal to the membranes:
    # u_n = ux_mem * cells.mem_vects_flat[:, 2] + uy_mem * cells.mem_vects_flat[:, 3]
    #
    # # calculate divergence as the sum of this vector x each surface area, divided by cell volume:
    # div_u = (np.dot(cells.M_sum_mems, u_n * cells.mem_sa) / cells.cell_vol)
    #
    #
    # calculate the reaction pressure required to counter-balance the flow field:
    # P_react = np.dot(cells.lapGJinv, div_u)
    #
    # else:
    #
    #     # calculate the reaction pressure required to counter-balance the flow field:
    #     P_react = np.dot(cells.lapGJinv, div_u)
    #
    # # calculate its gradient:
    # gradP_react = (P_react[cells.cell_nn_i[:, 1]] - P_react[cells.cell_nn_i[:, 0]]) / (cells.nn_len)
    #
    # gP_x = gradP_react * cells.mem_vects_flat[:,2]
    # gP_y = gradP_react * cells.mem_vects_flat[:,3]
    #
    # # average the components of the reaction force field at cell centres and get boundary values:
    # gPx_cell = np.dot(cells.M_sum_mems, gP_x) / cells.num_mems
    # gPy_cell = np.dot(cells.M_sum_mems, gP_y) / cells.num_mems
    #
    # # calculate the displacement of cell centres under the applied force under incompressible conditions:
    # sim.d_cells_x = sim.d_cells_x - gPx_cell
    # sim.d_cells_y = sim.d_cells_y - gPy_cell


    # if p.fixed_cluster_bound is True:  # enforce zero displacement boundary condition:
    #
    #     sim.d_cells_x[cells.bflags_cells] = 0
    #     sim.d_cells_y[cells.bflags_cells] = 0
    #
    #     # sim.d_cells_x[cells.nn_bound] = 0
    #     # sim.d_cells_y[cells.nn_bound] = 0
    #
    # else: # create a single fixed point on the boundary to prevent it from floating away
    #
    #     sim.d_cells_x[cells.bflags_cells[0]] = 0
    #     sim.d_cells_y[cells.bflags_cells[0]] = 0


    # check the displacement for NANs:
    stb.check_v(sim.d_cells_x)


def implement_deform_timestep(sim, cells, t, p):
    '''
    Implements the deformation of the tissue cluster based on divergence-free
    deformation calculated for cell centres.
    '''

    # create a smooth bivariate spline to interpolate deformation data from cells:
    cellinterp_x = SmoothBivariateSpline(cells.cell_centres[:, 0], cells.cell_centres[:, 1], sim.d_cells_x, kx=4, ky=4)
    cellinterp_y = SmoothBivariateSpline(cells.cell_centres[:, 0], cells.cell_centres[:, 1], sim.d_cells_y, kx=4, ky=4)

    # calculate deformations wrt the ecm using the smooth bivariate spline:
    dxv = cellinterp_x.ev(cells.mem_verts[:,0], cells.mem_verts[:,1])
    dyv = cellinterp_y.ev(cells.mem_verts[:,0], cells.mem_verts[:,1])

    xv2 = cells.mem_verts[:, 0] + dxv
    yv2 = cells.mem_verts[:, 1] + dyv

    # calculate new cell centres:
    cell_cent_x = np.dot(cells.M_sum_mems, xv2*cells.mem_sa)/cells.cell_sa
    cell_cent_y = np.dot(cells.M_sum_mems, yv2*cells.mem_sa)/cells.cell_sa

    # smooth the vertices:
    # xv2 = sim.smooth_weight_mem*xv2 + cell_cent_x[cells.mem_to_cells]*sim.smooth_weight_o
    # yv2 = sim.smooth_weight_mem*yv2 + cell_cent_y[cells.mem_to_cells]*sim.smooth_weight_o

    # repackage the vertices:
    cell_verts2 = []

    for i, _ in enumerate(cells.cell_verts):
        vx = xv2[cells.cell_to_mems[i]]
        vy = yv2[cells.cell_to_mems[i]]

        pts2 = np.column_stack((vx, vy))
        cell_verts2.append(pts2)

    cells.cell_verts = np.asarray(cell_verts2, dtype=object)
    cells.cell_centres = np.column_stack((cell_cent_x, cell_cent_y))

    # # calculate deformations wrt the ecm using the smooth bivariate spline:
    # decm_x = cellinterp_x.ev(cells.ecm_verts_unique[:, 0], cells.ecm_verts_unique[:, 1])
    # decm_y = cellinterp_y.ev(cells.ecm_verts_unique[:, 0], cells.ecm_verts_unique[:, 1])
    #
    # # get the new ecm verts by applying the deformation:
    # ecm_x2 = cells.ecm_verts_unique[:, 0] + decm_x
    # ecm_y2 = cells.ecm_verts_unique[:, 1] + decm_y
    #
    # ecm_new = np.column_stack((ecm_x2, ecm_y2))
    #
    # # repackage new ecm vertices as cells.ecm_verts:
    # ecm_verts2 = []
    #
    # for inds in cells.inds2ecmVerts:
    #     ecm_verts2.append(ecm_new[inds])

    # cells.ecm_verts = np.asarray(ecm_verts2)

    # rebuild essential portions of the cell world:
    # cells.deformWorld(p, ecm_verts2)

    # write data to time-storage vectors:
    sim.cell_centres_time.append(cells.cell_centres[:])
    # sim.mem_mids_time.append(cells.mem_mids_flat[:])
    # sim.maskM_time.append(cells.maskM[:])
    # sim.mem_edges_time.append(cells.mem_edges_flat[:])
    sim.cell_verts_time.append(cells.cell_verts[:])



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level package-specific facilities.

All functions defined by this submodule accept at least a previously imported
package object; most also accept the fully-qualified name of a package.

See Also
----------
:mod:`betse.util.py.module.pyname`
    Related submodule whose functions accept only fully-qualified names.
'''

# ....................{ IMPORTS                           }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable exceptions on missing mandatory
# dependencies, the top-level of this module may import *ONLY* from packages
# guaranteed to exist at installation time -- which typically means *ONLY*
# BETSE packages and stock Python packages.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

from betse.exceptions import BetsePackageException
# from betse.util.io.log import logs
from betse.util.type.types import (
    type_check, ModuleType, StrOrNoneTypes)

# ....................{ GETTERS ~ object                  }....................
@type_check
def get_object_type_package_root(obj: object) -> ModuleType:
    '''
    **Root package** (i.e., topmost package whose package name contains no
    ``.`` delimiters) transitively defining the class of the passed object.

    Parameters
    ----------
    obj : object
        Object to retrieve the root package name of.

    Returns
    ----------
    ModuleType
        Root package transitively defining this object's class.
    '''

    # Avoid circular import dependencies.
    from betse.util.py.module import pymodname

    # Fully-qualified name of the root package of this object's class.
    class_package_root_name = get_object_type_package_root_name(obj)

    # Return the previously imported instance of this root package.
    return pymodname.import_module(module_name=class_package_root_name)


@type_check
def get_object_type_package_root_name(obj: object) -> str:
    '''
    Name of the **root package** (i.e., topmost package whose package name
    contains no ``.`` delimiters) transitively defining the class of the passed
    object if any *or* raise an exception otherwise (i.e., if that class is
    defined by no module, as is the case for standard C-based classes).

    Design
    ----------
    The name of this function is intentionally suffixed by neither
    ``_qualified`` nor ``_unqualified``. Since root package names contain no
    ``.`` delimiters (e.g., :mod:`betse`), the fully-qualified and unqualified
    names for any root package are necessarily identical.

    Parameters
    ----------
    obj : object
        Object to retrieve the root package name of.

    Returns
    ----------
    str
        Name of the root package transitively defining this object's class.

    Raises
    ----------
    BetseTypeException
        If this object's class is *not* defined by a module (e.g., :mod:`str`).
        See the :func:`betse.util.type.cls.classes.get_module_name_qualified`
        function for further details.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.cls import classes
    from betse.util.type.obj import objects
    from betse.util.type.text.string import strs

    # Class of this object.
    cls = objects.get_class(obj)

    # Fully-qualified name of the module defining this class if any *OR* raise
    # the "BetseTypeException" otherwise.
    module_name = classes.get_module_name_qualified(cls)

    # Name of the root package defining this module. Note that:
    #
    # * If this module name is itself this package name, this module name
    #   contains no "." delimiter, in which case this call correctly returns
    #   this module name as is.
    # * Else, this module name contains at least one "." delimiter, in which
    #   case this call correctly returns the prefix of this module name
    #   preceding the first such ".".
    package_root_name = strs.get_prefix_preceding_char_or_text(
        text=module_name, char='.')

    # Return this name.
    return package_root_name

# ....................{ GETTERS ~ path                    }....................
@type_check
def get_package_project_dirname(package: ModuleType) -> str:
    '''
    **Absolute canonical dirname** (i.e., absolute dirname after resolving
    symbolic links) of the **root project directory** (i.e., top-level
    directory containing an installable ``pyproject.toml`` file or ``setup.py``
    script) governing the passed top-level Python package if found *or* raise
    an exception otherwise (i.e., if this directory is *not* found).

    Raises
    ----------
    BetsePackageException
        If no such directory exists for this package.

    See Also
    ----------
    :func:`get_package_project_dirname_or_none`
        Further details.
    '''

    # Avoid circular import dependencies.
    from betse.util.py.module import pymodule

    # Absolute canonical dirname of this package's project directory if found
    # *OR* "None" otherwise.
    package_project_dirname = get_package_project_dirname_or_none(package)

    # If this directory does *NOT* exist, raise an exception.
    if package_project_dirname is None:
        raise BetsePackageException(
            'Package "{}" project directory not found.'.format(
                pymodule.get_name_qualified(package)))
    # Else, this directory exists.

    # Return this dirname.
    return package_project_dirname


@type_check
def get_package_project_dirname_or_none(
    package: ModuleType) -> StrOrNoneTypes:
    '''
    **Absolute canonical dirname** (i.e., absolute dirname after resolving
    symbolic links) of the **root project directory** (i.e., top-level
    directory containing an installable ``pyproject.toml`` file or ``setup.py``
    script) governing the passed top-level Python package if found *or* raise
    an exception otherwise (i.e., if this directory is *not* found).

    Equivalently, this is the same as both:

    * The root directory archived by release tarballs for this application.
    * The Git-based working tree for this application (i.e., the top-level
      directory containing this application's ``.git`` subdirectory).

    Specifically, this function returns non-``None`` only if the parent
    directory of the passed package's directory contains either:

    * A ``pyproject.toml`` file, implying this package to be satisfy `PEP 518`_
      and hence be installable by at least poetry_.
    * A ``setup.py`` script, implying this package to be installable by either
      the standard :mod:`distutils` API or third-party :mod:`setuptools` API.

    :: _PEP 518:
       https://snarky.ca/clarifying-pep-518/
    :: _poetry:
       https://github.com/sdispater/poetry

    Caveats
    ----------
    **This directory typically does not exist.** This directory is only
    required during installation by non-developers *or* during development
    by developers. Once this application has been installed in a standard
    (i.e., non-editable) fashion by non-developers, this directory is no
    longer required and hence should *not* be assumed to exist.

    Parameters
    ----------
    package : ModuleType
        Top-level Python package to be queried.

    Returns
    ----------
    StrOrNoneTypes
        Either:

        * If the parent directory of the passed package's directory contains
          either a ``pyproject.toml`` or ``setup.py`` script, the absolute
          canonical dirname of that parent directory.
        * Else, ``None``.
    '''

    # Avoid circular import dependencies.
    from betse.util.path import files, pathnames
    from betse.util.py.module import pymodule

    # Absolute canonical dirname of the directory providing this package,
    # canonicalized into a directory rather than symbolic link to increase the
    # likelihood of obtaining the actual parent directory of this package.
    package_dirname = pymodule.get_dirname_canonical(package)

    # Absolute dirname of the parent directory of this directory.
    package_project_dirname = pathnames.get_dirname(package_dirname)

    # Absolute filenames of the "pyproject.toml" and "setup.py" files possibly
    # provided by this parent directory.
    pyproject_filename = pathnames.join(
        package_project_dirname, 'pyproject.toml')
    setup_filename = pathnames.join(package_project_dirname, 'setup.py')

    # Return this parent directory's absolute dirname if either of these
    # requisite installation-time files exist *OR* "None" otherwise.
    return (
        package_project_dirname
        if (
            files.is_file(pyproject_filename) or
            files.is_file(setup_filename)
        ) else
        None
    )



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **kernel** (i.e., central core software of the current operating
system) facilities.

Caveats
----------
**Operating system-specific logic is poor form.** Do so *only* where necessary.
'''

# ....................{ IMPORTS                           }....................
import platform
from betse.util.type.decorator.decmemo import func_cached
from betse.util.type.iterable.mapping.mapcls import OrderedArgsDict
from betse.util.type.types import type_check, VersionTypes

# ....................{ TESTERS                           }....................
@type_check
def is_version_greater_than_or_equal_to(version: VersionTypes) -> bool:
    '''
    ``True`` only if the current kernel version is greater than or equal to the
    passed version.

    Parameters
    ----------
    version : VersionTypes
        Version (e.g., ``1.0.2``, ``(1, 0, 2)``) to test the current kernel
        version against.

    Returns
    ----------
    bool
        ``True`` only if the current kernel version is greater than or equal to
        this version.

    Raises
    ----------
    pkg_resources.packaging.version.InvalidVersion
        If this version is *not* `PEP 440-compliant`_.

    .. _PEP 440-compliant:
       https://www.python.org/dev/peps/pep-0440

    See Also
    ----------
    :func:`get_version`
        Further details.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.numeric import versions

    # Current kernel version as a human-readable string.
    kernel_version = get_version()

    # Return true only if this version is greater than or equal to the passed
    # version.
    return versions.is_greater_than_or_equal_to(kernel_version, version)

# ....................{ GETTERS                           }....................
@func_cached
def get_name() -> str:
    '''
    Machine-readable name of the current kernel.

    While preserving human-readability is helpful, machine-readability is the
    principal use case for this function's return value (e.g., for use in
    dictionaries keyed on kernel type). This function returns:

    * Under Linux, ``Linux``.
    * Under macOS, ``Darwin``.
    * Under Windows (both Cygwin and vanilla), ``Windows``.
    * Under all other platforms, the string returned by the
      :func:`platform.system` function.
    '''

    # Return platform.system() as is, which appears to exactly corresponding to
    # the name of the current kernel on all platforms (e.g., "Darwin").
    return platform.system()


@func_cached
def get_version() -> str:
    '''
    Human-readable ``.``-delimited version specifier of the current kernel.

    This function returns:

    * Under Linux, the current Linux kernel version (e.g.,
      ``4.1.15-gentoo-r1``),
    * Under macOS, the current XNU kernel version (e.g., ``13.0.0``).
    * Under Windows, the current Windows kernel version (e.g., ``10.0.10240``).
    * Under all other platforms, the string returned by the
      :func:`platform.release` function.
    '''

    # Avoid circular import dependencies.
    from betse.util.os.brand import windows

    # Kernel version specifier to be returned.
    kernel_version = None

    # If the current platform is Windows, defer to the platform.version()
    # function. For only this platform, this function returns the desired
    # fine-grained Windows kernel version:
    #
    #     # Under Windows 10...
    #     >>> import platform
    #     >>> platform.release()
    #     "10"
    #     >>> platform.version()
    #     "10.0.10240"
    #     >>> platform.win32_ver()[1]
    #     "6.2.9200"
    #
    # However, note that the above does *NOT* generalize to any other platform:
    #
    #     # Under Gentoo Linux...
    #     >>> import platform
    #     >>> platform.release()
    #     "4.19.27-gentoo-r1"
    #     >>> platform.version()
    #     "#1 SMP Fri Apr 26 20:12:45 EDT 2019"
    #
    # Version specifier to be returned, defaulting to that returned by
    # platform.release(). While this typically corresponds to the low-level
    # version of the current kernel (e.g., "4.1.15" under Linux, "13.0.0" under
    # macOS), this is *NOT* necessarily the case (e.g., "8" rather than
    # "6.2.9200" under Windows). Hence, this is only a fallback.
    if windows.is_windows():
        kernel_version = platform.version()
    # Else, the current platform is *NOT* Windows. In this case, prefer the
    # version specifier returned by the platform.release() function to the
    # irrelevant timestamp returned by the platform.version() function.
    # Ironically, the platform.version() function does *NOT* actually return a
    # version specifier under most platforms. (Shaking my head.)
    else:
        kernel_version = platform.release()

    # Return this version.
    return kernel_version

# ....................{ GETTERS ~ metadata                }....................
def get_metadata() -> OrderedArgsDict:
    '''
    Ordered dictionary synopsizing the current kernel.
    '''

    # Return this dictionary.
    return OrderedArgsDict(
        'name', get_name(),
        'version', get_version(),
    )



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Gene regulatory network (GRN).

This submodule creates and electrodiffuses a suite of customizable general gene
products in the BETSE ecosystem, where the gene products are assumed to
activate and/or inhibit the expression of other genes (and therefore the
production of other gene products) in the gene regulatory network (GRN).
'''

#FIXME: Unify the large amount of code shared in common between this and the
#"gene" submodule as follows:
#
#* Define a new "grnabc" submodule defining a new "GrnABC" abstract base class,
#  defining at least:
#  * run_core_sim().
#  * All other methods shared in common between the "MasterOfGenes" and
#    "MasterOfMolecules" classes.
#* Refactor the "MasterOfGenes" and "MasterOfMolecules" classes to subclass the
#  "GrnABC" abstract base class.

# ....................{ IMPORTS                           }....................
import matplotlib.pyplot as plt
import numpy as np
from betse.science import filehandling as fh
from betse.science.chemistry.netplot import set_net_opts
from betse.science.chemistry.networks import MasterOfNetworks
from betse.science.organelles.microtubules import Mtubes
from betse.science.phase.phasecls import SimPhase
from betse.science.visual.plot import plotutil as viz
from betse.util.io.log import logs
from betse.util.type.types import type_check

# ....................{ CLASSES                           }....................
class MasterOfGenes(object):
    '''
    Gene regulatory network (GRN).
    '''

    # ..................{ INITIALIZERS                      }..................
    #FIXME: Let's try to remove this method altogether, if we can. Tree frogs!
    def __init__(self, p):
        pass


    @type_check
    def reinitialize(self, phase: SimPhase) -> None:

        # Localize high-level phase objects for convenience.
        cells = phase.cells
        p     = phase.p
        sim   = phase.sim

        # Previously loaded GRN-specific configuration file as a dictionary.
        config_dic = p.grn.conf

        # Time dilation:
        self.core.time_dila = float(config_dic.get('time dilation factor', 1.0))

        # reset microtubules?
        self.reset_MT = config_dic.get('reset microtubules', False)

        # recalculate fluid flow?
        self.recalc_fluid = config_dic.get('recalculate fluid flow', False)

        # obtain specific sub-dictionaries from the config file:
        substances_config = config_dic['biomolecules']
        reactions_config = config_dic.get('reactions', None)
        transporters_config = config_dic.get('transporters', None)
        channels_config = config_dic.get('channels', None)
        modulators_config = config_dic.get('modulators', None)

        self.core.tissue_init(phase, substances_config)

        if reactions_config is not None:
            # initialize the reactions of metabolism:
            self.core.read_reactions(reactions_config, sim, cells, p)
            self.core.write_reactions()
            self.core.create_reaction_matrix()
            self.core.write_reactions_env()
            self.core.create_reaction_matrix_env()

            self.reactions = True

        else:
            self.core.create_reaction_matrix()
            self.core.create_reaction_matrix_env()
            self.reactions = False

        # initialize transporters, if defined:
        if transporters_config is not None:
            self.core.read_transporters(transporters_config, phase)
            self.core.write_transporters(sim, cells, p)

            self.transporters = True

        else:
            self.transporters = False

        # initialize channels, if desired:
        if channels_config is not None:
            self.core.read_channels(channels_config, phase)
            self.channels = True

        else:
            self.channels = False

        # initialize modulators, if desired:
        if modulators_config is not None:
            self.core.read_modulators(modulators_config, sim, cells, p)
            self.modulators = True

        else:
            self.modulators = False

    # ..................{ READERS                           }..................
    @type_check
    def read_gene_config(self, phase: SimPhase) -> None:

        # Localize high-level phase objects for convenience.
        cells = phase.cells
        p     = phase.p
        sim   = phase.sim

        # Previously loaded GRN-specific configuration file as a dictionary.
        config_dic = p.grn.conf

        # obtain specific sub-dictionaries from the config file:
        substances_config = config_dic['biomolecules']
        reactions_config = config_dic.get('reactions', None)
        transporters_config = config_dic.get('transporters', None)
        channels_config = config_dic.get('channels', None)
        modulators_config = config_dic.get('modulators', None)

        # initialize the substances of metabolism in a core field encapsulating
        # Master of Molecules:
        self.core = MasterOfNetworks(sim, cells, substances_config, p)

        # Time dilation:
        self.core.time_dila = float(config_dic.get('time dilation factor', 1.0))

        # reset microtubules?
        self.reset_MT = config_dic.get('reset microtubules', False)

        # recalculate fluid flow?
        self.recalc_fluid = config_dic.get('recalculate fluid', False)

        # read in substance properties from the config file, and initialize basic properties:
        self.core.read_substances(sim, cells, substances_config, p)
        self.core.tissue_init(phase, substances_config)

        if reactions_config is not None:
            # initialize the reactions of metabolism:
            self.core.read_reactions(reactions_config, sim, cells, p)
            self.core.write_reactions()
            self.core.create_reaction_matrix()
            self.core.write_reactions_env()
            self.core.create_reaction_matrix_env()

            self.reactions = True
        else:
            self.core.create_reaction_matrix()
            self.core.create_reaction_matrix_env()
            self.reactions = False

        # initialize transporters, if defined:
        if transporters_config is not None:
            self.core.read_transporters(transporters_config, phase)
            self.core.write_transporters(sim, cells, p)

            self.transporters = True
        else:
            self.transporters = False

        # initialize channels, if desired:
        if channels_config is not None:
            self.core.read_channels(channels_config, phase)
            self.channels = True
        else:
            self.channels = False

        # initialize modulators, if desired:

        if modulators_config is not None:
            self.core.read_modulators(modulators_config, sim, cells, p)
            self.modulators = True
        else:
            self.modulators = False

        # read in network plotting options:
        self.core.net_plot_opts = config_dic.get('network plotting', None)

        # set plotting options for the network:
        set_net_opts(self.core, self.core.net_plot_opts, p)

        optim_exists = config_dic.get('optimization', None)
        if optim_exists is not None:

            # after primary initialization, check and see if optimization required:
                # after primary initialization, check and see if optimization required:
            opti = config_dic['optimization']['optimize network']
            self.core.opti_N = config_dic['optimization']['optimization steps']
            self.core.opti_method = config_dic['optimization']['optimization method']
            self.core.target_vmem = float(config_dic['optimization']['target Vmem'])
            self.core.opti_T = float(config_dic['optimization']['optimization T'])
            self.core.opti_step = float(config_dic['optimization']['optimization step'])
            # self.core.opti_run = config_dic['optimization']['run from optimization']

            if opti:
                logs.log_info('Analyzing gene network for optimal rates...')
                self.core.optimizer(sim, cells, p)
                self.reinitialize(phase)

    # ..................{ RUNNERS                           }..................
    @type_check
    def run_core_sim(self, phase: SimPhase) -> None:
        '''
        Simulate this gene regulatory network (GRN) in an isolated manner
        ignoring all other biophysicality (e.g., bioelectricity, fluid flow).

        Parameters
        ----------
        phase : SimPhase
            Current simulation phase.
        '''

        # Localize high-level phase objects for convenience.
        cells = phase.cells
        p     = phase.p
        sim   = phase.sim

        # set molecules to not affect charge for sim-grn test-drives:
        p.substances_affect_charge = False

        #FIXME: This... is quite unfortunate. Ideally, core parameters in the
        #"Parameters" object should *NEVER* be modified, as doing so actually
        #modifies the underlying YAML dictionary structure. If the current
        #simulation configuration file is saved to disk, the user-defined
        #values for these parameters will be overwritten by the new hacky
        #values defined below. Which would be bad. Is there no alternative?

        # Set sim-grn specific time step and sampling:
        p.dt = p.grn_dt
        p.init_time_total = p.grn_total_time
        p.init_time_sampling = p.grn_tsample

        p.init_tsteps = int(p.init_time_total / p.dt)
        p.resample = p.init_time_sampling
        p.total_time = p.init_time_total

        # Number of time steps (including sampled and unsampled) between each
        # unsampled time step, including that unsampled time step itself.
        p.t_resample = p.resample / p.dt

        # specify a time vector
        loop_time_step_max = p.init_tsteps
        # Maximum number of seconds simulated by the current run.
        loop_seconds_max = loop_time_step_max * p.dt
        # Time-steps vector appropriate for the current run.
        tt = np.linspace(0, loop_seconds_max, loop_time_step_max)

        tsamples = set()
        i = 0
        while i < len(tt) - p.t_resample:
            i = int(i + p.t_resample)
            # logs.log_debug('Time sample i: {!r}'.format(i))
            tsamples.add(tt[i])

        # if p.grn_runmodesim:
        self.reinitialize(phase)

        self.core.clear_cache()
        self.time = []

        self.mod_after_cut = False # set this to false

        if self.recalc_fluid:  # If user requests the GRN recalculate/calculate fluid:
            logs.log_info("Calculating fluid in terms of endogenous currents...")

            p.fluid_flow = True  # turn fluid flow on (in case it was off)

            # calculate or re-calculate fluid flow in terms of curl-component of current:
            cc = sim.cc_env.mean(axis=0).reshape(cells.X.shape)
            zz = sim.zs.mean()

            sim.u_env_x = -sim.J_env_x / (p.F * cc * zz)
            sim.u_env_y = -sim.J_env_y / (p.F * cc * zz)

        # if p.use_microtubules:
        #     sim.mtubes.reinit(cells, p)
        #
        #     self.mtubes_x_time = []
        #     self.mtubes_y_time = []
        #
        #     if self.reset_MT:
        #         logs.log_info(
        #             'Resetting microtubules for sim-grn simulation...')
        #         sim.mtubes = Mtubes(sim, cells, p)

        for t in tt:
            if self.transporters:
                self.core.run_loop_transporters(t, sim, cells, p)

            self.core.run_loop(phase=phase, t=t)

            # if p.use_microtubules: # update the microtubules:
            #     sim.mtubes.update_mtubes(cells, sim, p)

            # If...
            if (
                # The simulation phase is being run...
                p.grn_runmodesim and
                # The cutting event has yet to be performed here...
                not self.mod_after_cut and
                # The current time step is at least the time step at which the
                # cutting event is scheduled to occur...
                t >= p.event_cut_time
            ):
                phase.dyna.fire_events(phase=phase, t=t)

                #FIXME: This logic should ideally reside somewhere in the
                #"betse.science.tissue" subpackage. For example, similar logic
                #already exists in the
                #betse.science.tissue.tishandler.TissueHandler._sim_events_tissue()
                #method called by the above phase.dyna.fire_events() call. It
                #should thus be possible to shift the body of the following if
                #conditional into the body of the similar if conditional in the
                #TissueHandler._sim_events_tissue() method. After doing so,
                #this outer if conditional and the "mod_after_cut" variable
                #would then be safely removable. Overlord, unite your power!

                # If a cutting event has just been run...
                if (phase.dyna.event_cut.is_fired and
                    not self.mod_after_cut):
                    self.core.mod_after_cut_event(
                        phase,
                        sim.target_inds_cell_o,
                        sim.target_inds_mem_o,
                    )
                    self.core.redefine_dynamic_dics(sim, cells, p)

                    logs.log_info(
                        "Reinitializing the gene regulatory network for simulation...")
                    self.reinitialize(phase)
                    # self.core.clear_cache()
                    # sim.uxmt, self.uymt = sim.mtubes.mtubes_to_cell(cells, p)

                    self.mod_after_cut = True  # set the boolean to avoid repeat action

            if t in tsamples:
                sim.time.append(t)

                logs.log_info('------------------' + str(np.round(t,3)) +' s --------------------')
                self.time.append(t)
                self.core.write_data(sim, cells, p)
                self.core.report(sim, p)

                # if p.use_microtubules:
                #     # microtubules:
                #     self.mtubes_x_time.append(sim.mtubes.mtubes_x * 1)
                #     self.mtubes_y_time.append(sim.mtubes.mtubes_y * 1)

        logs.log_info('Saving simulation...')
        datadump = [self, cells, p]
        fh.saveSim(p.grn_pickle_filename, datadump)
        self.core.init_saving(cells, p, plot_type='grn', nested_folder_name='RESULTS')

        # # microtubules plot------------------------------------------------------------------------
        # if p.use_microtubules:
        #
        #     logs.log_info("Plotting microtubules used in GRN simulation...")
        #
        #     plt.figure()
        #     ax = plt.subplot(111)
        #
        #     umtx, umty = sim.mtubes.mtubes_to_cell(cells, p)
        #
        #     plt.figure()
        #     ax = plt.subplot(111)
        #
        #     viz.plotVectField(
        #         umtx,
        #         umty,
        #         cells,
        #         p,
        #         plot_ecm=False,
        #         title='Final Microtubule Alignment Field',
        #         cb_title='Aligned MT Fraction',
        #         colorAutoscale=False,
        #         minColor=0.0,
        #         maxColor=1.0,
        #     )
        #
        #     # viz.mem_quiver(
        #     #     sim.mtubes.mtubes_x,
        #     #     sim.mtubes.mtubes_y,
        #     #     ax,
        #     #     cells,
        #     #     p,
        #     # )
        #
        #     ax.set_xlabel('X-Distance [um]')
        #     ax.set_ylabel('Y-Distance [um]')
        #     ax.set_title('Microtubule arrangement in cells')
        #
        #     if p.autosave is True:
        #         savename = self.core.imagePath + 'Microtubules' + '.png'
        #         plt.savefig(savename, format='png', transparent=True)
        #
        #     if p.turn_all_plots_off is False:
        #         plt.show(block=False)

        if self.recalc_fluid:

            logs.log_info("Plotting fluid flow used in GRN simulation")

            plt.figure()
            ax = plt.subplot(111)

            viz.plotStreamField(
                1e9 * sim.u_env_x,
                1e9 * sim.u_env_y,
                cells, p,
                plot_ecm=True,
                title='Final Fluid Velocity in Environment',
                cb_title='Velocity [nm/s]'
            )

            if p.autosave is True:
                savename = self.core.imagePath + 'Fluid_ECM' + '.png'
                plt.savefig(savename, format='png', transparent=True)

            if p.turn_all_plots_off is False:
                plt.show(block=False)


        self.core.export_eval_strings(p)
        self.core.export_equations(p)
        message = 'Gene regulatory network simulation saved to' + ' ' + p.grn_pickle_filename
        logs.log_info(message)

        logs.log_info('-------------------Simulation Complete!-----------------------')



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level object facilities.
'''

# ....................{ IMPORTS                           }....................
from betse.exceptions import BetseAttrException, BetseCallableException
from betse.util.type.types import (
    type_check,
    CallableTypes,
    CallableOrNoneTypes,
    ClassType,
    TestableOrNoneTypes,
)

# ....................{ GETTERS : attr                    }....................
@type_check
def get_attr(obj: object, attr_name: str, **kwargs) -> object:
    '''
    Value of the attribute with the passed name bound to the passed object if
    this object defines this attribute *or* raise an exception otherwise (i.e.,
    if this object defines no such attribute), optionally validated to be of
    the passed type.

    Parameters
    ----------
    obj : object
        Object to be inspected.
    attr_name : str
        Name of the attribute to be retrieved.

    All remaining keyword arguments are passed as is to the
    :func:`get_attr_or_default` function.

    Returns
    ----------
    object
        Value of the attribute with this name bound to this object.

    Raises
    ----------
    BetseAttrException
        If no such attribute is bound to this object.
    BetseTypeException
        If the ``attr_type`` parameter is non-``None`` and the type of the
        current value of this attribute is *not* an instance of ``attr_type``.
    '''

    # Avoid circular import dependencies.
    from betse.util.py.module import pymodule
    from betse.util.type.obj.sentinels import SENTINEL

    # Value of the attribute with this name defined by this object if any *OR*
    # the sentinel otherwise.
    attr_value = get_attr_or_sentinel(obj, attr_name, **kwargs)

    # If no such attribute exists...
    if attr_value is SENTINEL:
        # Human-readable name of this object. Specifically:
        #
        # * If this object is a module, the fully-qualified name of this
        #   module (e.g., "betse.metadeps").
        # * Else, the unqualified name of the class of this object.
        #
        # Since the unqualified name of the class of module objects is simply
        # "module", differentiating these two common cases improves sanity.
        obj_name = (
            pymodule.get_name_qualified(obj) if pymodule.is_module(obj) else
            get_class_name_unqualified(obj))

        # Raise an exception embedding this object name.
        raise BetseAttrException(
            'Attribute "{}.{}" undefined.'.format(obj_name, attr_name))
    # Else, this attribute exists.

    # Return this value.
    return attr_value


@type_check
def get_attr_or_none(obj: object, attr_name: str, **kwargs) -> object:
    '''
    Attribute with the passed name bound to the passed object if this object
    defines this attribute *or* ``None`` otherwise (i.e., if this object
    defines no such attribute), optionally validated to be of
    the passed type.

    Caveats
    ----------
    For disambiguity, consider calling the :func:`get_attr_or_sentinel`
    function instead. Whereas this function fails to distinguish between
    existing attributes whose values are ``None`` and non-existing attributes
    for which this function returns ``None``, the :func:`get_attr_or_sentinel`
    function trivially disambiguates between these two common edge cases.

    Parameters
    ----------
    obj : object
        Object to be inspected.
    attr_name : str
        Name of the attribute to be obtained.

    All remaining keyword arguments are passed as is to the
    :func:`get_attr_or_default` function.

    Returns
    ----------
    object
        Attribute with this name bound to this object if any *or* ``None``
        otherwise.

    Raises
    ----------
    BetseTypeException
        If the ``attr_type`` parameter is non-``None`` and the type of the
        current value of this attribute is *not* an instance of ``attr_type``.
    '''

    # Return the current value of the attribute with this name defined by this
    # object if any *OR* "None" otherwise.
    return get_attr_or_default(
        obj=obj, attr_name=attr_name, attr_default=None, **kwargs)


@type_check
def get_attr_or_sentinel(obj: object, attr_name: str, **kwargs) -> object:
    '''
    Value of the attribute with the passed name bound to the passed object if
    this object defines this attribute *or* the sentinel singleton otherwise
    (i.e., if this object defines no such attribute), optionally validated to
    be of the passed type.

    This function enables callers to safely distinguish between non-existing
    attributes and existing attributes whose values are ``None``.

    Parameters
    ----------
    obj : object
        Object to be inspected.
    attr_name : str
        Name of the attribute to be obtained.

    All remaining keyword arguments are passed as is to the
    :func:`get_attr_or_default` function.

    Returns
    ----------
    object
        Either:

        * If this object declares this attribute, this attribute's value.
        * Else, the **sentinel singleton** (i.e.,
          :attr:`betse.util.type.obj.sentinels.SENTINEL`).

    Raises
    ----------
    BetseTypeException
        If the ``attr_type`` parameter is non-``None`` and the type of the
        current value of this attribute is *not* an instance of ``attr_type``.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.obj.sentinels import SENTINEL

    # Return the current value of the attribute with this name defined by this
    # object if any *OR* the sentinel otherwise.
    return get_attr_or_default(
        obj=obj, attr_name=attr_name, attr_default=SENTINEL, **kwargs)


@type_check
def get_attr_or_default(
    # Mandatory parameters.
    obj: object,
    attr_name: str,
    attr_default: object,

    # Optional parameters.
    attr_type: TestableOrNoneTypes = None,
) -> object:
    '''
    Value of the attribute with the passed name bound to the passed object if
    this object defines this attribute *or* the passed default value otherwise
    (i.e., if this object defines no such attribute), optionally validated to
    be of the passed type.

    Parameters
    ----------
    obj : object
        Object to be inspected.
    attr_name : str
        Name of the attribute to return the current value of.
    attr_default : object
        Default value to be returned if this object defines no such attribute.
    attr_type : TestableOrNoneTypes
        Expected type of the current value of this attribute. This function
        effectively performs the equivalent of the :meth:`type_check` decorator
        at runtime by raising an exception if all of the following apply:

        * This type is *not* ``None``.
        * This value is *not* this default value, implying this attribute to be
          defined by this object.
        * This value is *not* an instance of this type.

        Defaults to ``None``, in which case no such type checking is performed.

    Returns
    ----------
    object
        Either:

        * If this object defines this attribute, this attribute's value.
        * Else, this default value.

    Raises
    ----------
    BetseTypeException
        If the ``attr_type`` parameter is non-``None`` and the type of the
        current value of this attribute is *not* an instance of ``attr_type``.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.obj import objtest

    # Value of the attribute with this name defined by this object if any *OR*
    # this default value otherwise.
    attr_value = getattr(obj, attr_name, attr_default)

    # If this value is to be type-checked *AND* is *NOT* this default value
    # (which by definition already satisfies caller requirements regardless of
    # type), type-check this value.
    if attr_type is not None and attr_value is not attr_default:
        objtest.die_unless_instance(obj=attr_value, cls=attr_type)

    # Return this value.
    return attr_value

# ....................{ GETTERS : callable                }....................
@type_check
def get_callable(obj: object, callable_name: str) -> CallableTypes:
    '''
    Callable with the passed name defined by the passed object if this object
    defines such a callable *or* raise an exception otherwise (i.e., if this
    object defines no such callable).

    Parameters
    ----------
    obj : object
        Object to be queried.
    callable_name : str
        Name of the callable to be returned.

    Returns
    ----------
    CallableTypes
        Callable with this name defined by this object.

    Raises
    ----------
    BetseCallableException
        If this object defines no callable with this name.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.obj.sentinels import SENTINEL

    # Attribute with this name defined by this object if any *OR* the sentinel.
    # To raise human-readable exceptions, this attribute is *NOT* retrieved via
    # the higher-level get_callable_or_none() method; doing so would obscure
    # whether this attribute does not exist or does but is not callable.
    func = get_attr_or_sentinel(obj, callable_name)

    # If this attribute does *NOT* exist, raise an exception.
    if func is SENTINEL:
        raise BetseCallableException(
            'Method {}.{}() undefined.'.format(
                obj.__class__.__name__, callable_name))

    # If this attribute exists but is *NOT* callable, raise an exception.
    if not callable(func):
        raise BetseCallableException(
            'Object attribute "{}.{}" not callable: {!r}'.format(
                obj.__class__.__name__, callable_name, func))
    # Else, this attribute is callable.

    # Return this callable.
    return func


@type_check
def get_callable_or_none(obj: object, callable_name: str) -> (
    CallableOrNoneTypes):
    '''
    Callable with the passed name defined by the passed object if this object
    defines such a callable *or* ``None`` otherwise (i.e., if this object
    defines no such callable).

    Parameters
    ----------
    obj : object
        Object to be queried.
    callable_name : str
        Name of the callable to be returned.

    Returns
    ----------
    CallableOrNoneTypes
        Either:

        * If this object defines this callable, this callable.
        * Else, ``None``.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.obj.sentinels import SENTINEL

    # Attribute with this name defined by this object if any *OR* the sentinel.
    func = get_attr_or_sentinel(obj, callable_name)

    # Return this attribute if this attribute is callable *OR* "None".
    return func if func is not SENTINEL and callable(func) else None

# ....................{ GETTERS ~ class                   }....................
def get_class(obj: object) -> ClassType:
    '''
    Passed object if this object is itself a class *or* the class of this
    object otherwise (i.e., if this object is *not* a class).

    Parameters
    ----------
    obj : object
        Object to be queried for its class.

    Returns
    ----------
    ClassType
        This object if this object is a class *or* this object's class.
    '''

    # Simplicity is not a place in Simple City.
    return obj if isinstance(obj, ClassType) else type(obj)

# ....................{ GETTERS ~ class : name            }....................
def get_class_name_unqualified(obj: object) -> str:
    '''
    Unqualified name of either the passed object if this object is itself a
    class *or* the class of this object otherwise (i.e., if this object is
    *not* a class).

    Parameters
    ----------
    obj : object
        Object to be queried for its class name.

    Returns
    ----------
    str
        Unqualified name of this class.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.cls import classes

    # This object if this object is a class *OR* this object's class otherwise.
    cls = get_class(obj)

    # Return the unqualified name of this class.
    return classes.get_name_unqualified(cls)


def get_class_module_name_qualified(obj: object) -> str:
    '''
    Fully-qualified name of the module defining either the passed object if
    this object is itself a class *or* the class of this object otherwise
    (i.e., if this object is *not* a class).

    Parameters
    ----------
    obj : object
        Object to be queried for its module name.

    Returns
    ----------
    str
        Fully-qualified name of this module.

    Raises
    ----------
    BetseTypeException
        If this class has no ``__module__`` attribute, which should ideally
        *never* happen.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.cls import classes

    # This object if this object is a class *OR* this object's class otherwise.
    cls = get_class(obj)

    # Return the fully-qualified name of this class.
    return classes.get_module_name_qualified(cls)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Abstract mixins of YAML-backed configuration subclasses, standardizing property
nomenclature for common YAML design patterns.
'''

# ....................{ IMPORTS                           }....................
from betse.lib.yaml.yamlalias import yaml_alias
# from betse.util.type.types import type_check, ClassType, SequenceTypes

# ....................{ MIXINS ~ alias                    }....................
# Mixins standardizing various general-purpose YAML aliases.

class YamlBooledMixin(object):
    '''
    Mixin of all **YAML-backed booled configuration** (i.e., backed by a YAML
    dictionary with top-level key ``enabled`` whose value is a boolean
    specifying whether this configuration is enabled or disabled) subclasses.

    This class is suitable for use as a multiple-inheritance mixin. To preserve
    the expected method resolution order (MRO) semantics, this class should
    typically be inherited *first* rather than *last* in subclasses.

    Attributes
    ----------
    is_enabled : bool
        ``True`` only if this list item is enabled.
    '''

    # ..................{ ALIASES                           }..................
    is_enabled = yaml_alias("['enabled']", bool)


class YamlNamedMixin(object):
    '''
    Mixin of all **YAML-backed named configuration** (i.e., backed by a YAML
    dictionary with top-level key ``name`` whose value is a human- and/or
    machine-readable string identifying this configuration's presumably unique
    name) subclasses.

    This class is suitable for use as a multiple-inheritance mixin. To preserve
    the expected method resolution order (MRO) semantics, this class should
    typically be inherited *first* rather than *last* in subclasses.

    Attributes
    ----------
    name : str
        Arbitrary string typically uniquely identifying this configuration
        (e.g., ``spot``, naming a spatially isolated tissue profile).
    '''

    # ..................{ ALIASES                           }..................
    name = yaml_alias("['name']", str)


class YamlTypedMixin(object):
    '''
    Mixin of all **YAML-backed typed configuration** (i.e., backed by a YAML
    dictionary with top-level key ``type`` whose value is a machine-readable
    string identifying this configuration's type) subclasses.

    This class is suitable for use as a multiple-inheritance mixin. To preserve
    the expected method resolution order (MRO) semantics, this class should
    typically be inherited *first* rather than *last* in subclasses.

    Attributes
    ----------
    kind : str
        Lowercase alphanumeric string uniquely identifying the type of this
        configuration (e.g., ``voltage_membrane``, identifying a transmembrane
        voltage). See each ``type`` key of the corresponding list in the
        default simulation configuration file for real-world examples.
    '''

    # ..................{ ALIASES                           }..................
    kind = yaml_alias("['type']", str)
